import pytest

import salt.modules.useradd as useradd
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch


@pytest.fixture
def configure_loader_modules():
    return {
        useradd: {
            "__grains__": {
                "kernel": "Linux",
                "osarch": "x86_64",
                "os": "CentOS",
                "os_family": "RedHat",
                "osmajorrelease": 8,
            },
            "__salt__": {},
        }
    }


def test_add():
    # command found and successful run
    mock = MagicMock(return_value={"retcode": 0})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/useradd")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.add("Salt") is True
    mock.assert_called_once_with(["/sbin/useradd", "-m", "Salt"], python_shell=False)

    # command found and unsuccessful run
    mock = MagicMock(return_value={"retcode": 1})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/useradd")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.add("Salt") is False
    mock.assert_called_once_with(["/sbin/useradd", "-m", "Salt"], python_shell=False)

    # command not found
    mock = MagicMock()
    with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.dict(
        useradd.__salt__, {"cmd.run_all": mock}
    ):
        with pytest.raises(CommandExecutionError):
            useradd.add("Salt")
    mock.assert_not_called()


def test_delete():
    # command found and successful run
    mock = MagicMock(return_value={"retcode": 0})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/userdel")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.delete("Salt") is True
    mock.assert_called_once_with(["/sbin/userdel", "Salt"], python_shell=False)

    # command found and unsuccessful run
    mock = MagicMock(return_value={"retcode": 1})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/userdel")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.delete("Salt") is False
    mock.assert_called_once_with(["/sbin/userdel", "Salt"], python_shell=False)

    # command not found
    mock = MagicMock()
    with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.dict(
        useradd.__salt__, {"cmd.run_all": mock}
    ):
        with pytest.raises(CommandExecutionError):
            useradd.delete("Salt")
    mock.assert_not_called()


def test_chgroups():
    # groups matched - no command run
    mock = MagicMock()
    with patch.object(
        useradd, "list_groups", MagicMock(return_value=["wheel", "root"])
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.chgroups("Salt", "wheel,root") is True
    mock.assert_not_called()

    # command found and successful run
    mock = MagicMock(return_value={"retcode": 0})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.chgroups("Salt", "wheel,root") is True
    mock.assert_called_once_with(
        ["/sbin/usermod", "-G", "wheel,root", "Salt"], python_shell=False
    )

    # command found and unsuccessful run
    mock = MagicMock(return_value={"retcode": 1, "stderr": ""})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        assert useradd.chgroups("Salt", "wheel,root") is False
    mock.assert_called_once_with(
        ["/sbin/usermod", "-G", "wheel,root", "Salt"], python_shell=False
    )

    # command not found
    mock = MagicMock()
    with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.dict(
        useradd.__salt__, {"cmd.run_all": mock}
    ):
        with pytest.raises(CommandExecutionError):
            useradd.chgroups("Salt", "wheel,root")
    mock.assert_not_called()


def test_chloginclass():
    # only runs on OpenBSD
    assert useradd.chloginclass("Salt", "staff") is False

    with patch.dict(useradd.__grains__, {"kernel": "OpenBSD"}):
        # command found and successful run
        userinfo = ["class salt", "class staff"]
        mock = MagicMock(return_value={"retcode": 0})
        with patch(
            "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
        ), patch.dict(
            useradd.__salt__, {"cmd.run_stdout": MagicMock(side_effect=userinfo)}
        ), patch.dict(
            useradd.__salt__, {"cmd.run": mock}
        ):
            assert useradd.chloginclass("Salt", "staff") is True
        mock.assert_called_once_with(
            ["/sbin/usermod", "-L", "staff", "Salt"], python_shell=False
        )

        # command found and unsuccessful run
        userinfo = ["class salt", "class salt"]
        mock = MagicMock(return_value={"retcode": 1, "stderr": ""})
        with patch(
            "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
        ), patch.dict(
            useradd.__salt__, {"cmd.run_stdout": MagicMock(side_effect=userinfo)}
        ), patch.dict(
            useradd.__salt__, {"cmd.run": mock}
        ):
            assert useradd.chloginclass("Salt", "staff") is False
        mock.assert_called_once_with(
            ["/sbin/usermod", "-L", "staff", "Salt"], python_shell=False
        )

        # command not found
        userinfo = ["class salt"]
        mock = MagicMock()
        with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.dict(
            useradd.__salt__, {"cmd.run_stdout": MagicMock(side_effect=userinfo)}
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chloginclass("Salt", "staff")
        mock.assert_not_called()


def test__chattrib():
    # command found and successful run
    mock = MagicMock(return_value={"retcode": 0})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.object(
        useradd, "info", MagicMock(side_effect=[{"uid": 10}, {"uid": 11}])
    ), patch.dict(
        useradd.__salt__, {"cmd.run": mock}
    ):
        assert useradd._chattrib("Salt", "uid", 11, "-u") is True
    mock.assert_called_once_with(
        ["/sbin/usermod", "-u", 11, "Salt"], python_shell=False
    )

    # command found and unsuccessful run
    mock = MagicMock(return_value={"retcode": 1})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.object(
        useradd, "info", MagicMock(side_effect=[{"uid": 10}, {"uid": 10}])
    ), patch.dict(
        useradd.__salt__, {"cmd.run": mock}
    ):
        assert useradd._chattrib("Salt", "uid", 11, "-u") is False
    mock.assert_called_once_with(
        ["/sbin/usermod", "-u", 11, "Salt"], python_shell=False
    )

    # command not found
    mock = MagicMock()
    with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.object(
        useradd, "info", MagicMock(return_value={"uid": 10})
    ), patch.dict(useradd.__salt__, {"cmd.run_all": mock}):
        with pytest.raises(CommandExecutionError):
            useradd._chattrib("Salt", "uid", 11, "-u")
    mock.assert_not_called()


def test__update_gecos():
    pre_info = {"fullname": "Uli Kunkel"}
    post_info = {"fullname": "Karl Hungus"}

    # command found and successful run
    mock = MagicMock(return_value={"retcode": 0})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.object(
        useradd, "_get_gecos", MagicMock(side_effect=[pre_info, post_info])
    ), patch.dict(
        useradd.__salt__, {"cmd.run": mock}
    ):
        assert useradd._update_gecos("Salt", "fullname", post_info["fullname"]) is True
    mock.assert_called_once_with(
        ["/sbin/usermod", "-c", "Karl Hungus", "Salt"], python_shell=False
    )

    # command found and unsuccessful run
    mock = MagicMock(return_value={"retcode": 1})
    with patch(
        "salt.utils.path.which", MagicMock(return_value="/sbin/usermod")
    ), patch.object(
        useradd, "_get_gecos", MagicMock(side_effect=[pre_info, pre_info])
    ), patch.dict(
        useradd.__salt__, {"cmd.run": mock}
    ):
        assert useradd._update_gecos("Salt", "fullname", post_info["fullname"]) is False
    mock.assert_called_once_with(
        ["/sbin/usermod", "-c", "Karl Hungus", "Salt"], python_shell=False
    )

    # command not found
    mock = MagicMock()
    with patch("salt.utils.path.which", MagicMock(return_value=None)), patch.object(
        useradd, "_get_gecos", MagicMock(side_effect=[pre_info, pre_info])
    ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
        with pytest.raises(CommandExecutionError):
            useradd._update_gecos("Salt", "fullname", post_info["fullname"])
    mock.assert_not_called()


def test_rename():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "info", MagicMock(return_value={"uid": 10})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.rename("salt", 1)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "info", mock):
            with pytest.raises(CommandExecutionError):
                useradd.rename("salt", 1)

        mock = MagicMock(return_value=True)
        with patch.object(useradd, "info", mock):
            with pytest.raises(CommandExecutionError):
                useradd.rename("salt", 1)

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[False, {"name": ""}, {"name": "salt"}])
            with patch.object(useradd, "info", mock):
                assert useradd.rename("name", "salt") is True

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[False, {"name": ""}, {"name": ""}])
            with patch.object(useradd, "info", mock):
                assert useradd.rename("salt", "salt") is False


def test_chuid():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "info", MagicMock(return_value={"uid": 10})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chuid("salt", 1)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value={"uid": 11})
        with patch.object(useradd, "info", mock):
            assert useradd.chuid("name", 11) is True

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"uid": 11}, {"uid": 11}])
            with patch.object(useradd, "info", mock):
                assert useradd.chuid("name", 22) is False

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"uid": 11}, {"uid": 22}])
            with patch.object(useradd, "info", mock):
                assert useradd.chuid("name", 11) is True


def test_chgid():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "info", MagicMock(return_value={"gid": 10})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chgid("salt", 1)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value={"gid": 11})
        with patch.object(useradd, "info", mock):
            assert useradd.chgid("name", 11) is True

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"gid": 22}, {"gid": 22}])
            with patch.object(useradd, "info", mock):
                assert useradd.chgid("name", 11) is False

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(side_effect=[{"gid": 11}, {"gid": 22}])
            with patch.object(useradd, "info", mock):
                assert useradd.chgid("name", 11) is True


def test_chshell():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "info", MagicMock(return_value={"shell": "/bin/bash"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chshell("salt", "/usr/bash")
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value={"shell": "/bin/bash"})
        with patch.object(useradd, "info", mock):
            assert useradd.chshell("name", "/bin/bash") is True

        mock_run = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(
                side_effect=[{"shell": "/bin/bash"}, {"shell": "/bin/bash"}]
            )
            with patch.object(useradd, "info", mock):
                assert useradd.chshell("name", "/usr/bash") is False

        with patch.dict(useradd.__salt__, {"cmd.run": mock_run}):
            mock = MagicMock(
                side_effect=[{"shell": "/bin/bash"}, {"shell": "/usr/bash"}]
            )
            with patch.object(useradd, "info", mock):
                assert useradd.chshell("name", "/bin/bash") is True


def test_chhome():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "info", MagicMock(return_value={"home": "/root"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chhome("salt", "/user")
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value={"home": "/root"})
        with patch.object(useradd, "info", mock):
            assert useradd.chhome("name", "/root") is True

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[{"home": "/root"}, {"home": "/root"}])
            with patch.object(useradd, "info", mock):
                assert useradd.chhome("name", "/user") is False

        mock = MagicMock(return_value=None)
        with patch.dict(useradd.__salt__, {"cmd.run": mock}):
            mock = MagicMock(side_effect=[{"home": "/root"}, {"home": "/root"}])
            with patch.object(useradd, "info", mock):
                assert useradd.chhome("name", "/root") is True


def test_chfullname():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "_get_gecos", MagicMock(return_value={"fullname": "Salt"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chfullname("salt", "Saltstack")
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chfullname("Salt", "SaltStack") is False

        mock = MagicMock(return_value={"fullname": "SaltStack"})
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chfullname("Salt", "SaltStack") is True

        mock = MagicMock(return_value={"fullname": "SaltStack"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"fullname": "SaltStack2"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chfullname("Salt", "SaltStack1") is False

        mock = MagicMock(return_value={"fullname": "SaltStack2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"fullname": "SaltStack2"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chfullname("Salt", "SaltStack1") is False


def test_chroomnumber():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "_get_gecos", MagicMock(return_value={"roomnumber": "1"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chroomnumber("salt", 2)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chroomnumber("salt", 1) is False

        mock = MagicMock(return_value={"roomnumber": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chroomnumber("salt", 1) is True

        mock = MagicMock(return_value={"roomnumber": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"roomnumber": "3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chroomnumber("salt", 1) is False

        mock = MagicMock(return_value={"roomnumber": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"roomnumber": "3"})


def test_chworkphone():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "_get_gecos", MagicMock(return_value={"workphone": "1"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chworkphone("salt", 2)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chworkphone("salt", 1) is False

        mock = MagicMock(return_value={"workphone": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chworkphone("salt", 1) is True

        mock = MagicMock(return_value={"workphone": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"workphone": "3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chworkphone("salt", 1) is False

        mock = MagicMock(return_value={"workphone": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"workphone": "3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chworkphone("salt", 1) is False


def test_chhomephone():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "_get_gecos", MagicMock(return_value={"homephone": "1"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chhomephone("salt", 2)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chhomephone("salt", 1) is False

        mock = MagicMock(return_value={"homephone": "1"})
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chhomephone("salt", 1) is True

        mock = MagicMock(return_value={"homephone": "2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"homephone": "3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chhomephone("salt", 1) is False

        mock = MagicMock(return_value={"homephone": "3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"homephone": "3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chhomephone("salt", 1) is False


def test_chother():
    # command not found
    with patch("salt.utils.path.which", MagicMock(return_value=None)):
        mock = MagicMock()
        with patch.object(
            useradd, "_get_gecos", MagicMock(return_value={"other": "1"})
        ), patch.dict(useradd.__salt__, {"cmd.run": mock}):
            with pytest.raises(CommandExecutionError):
                useradd.chother("salt", 2)
        mock.assert_not_called()

    # command found
    with patch("salt.utils.path.which", MagicMock(return_value="/sbin/usermod")):
        mock = MagicMock(return_value=False)
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chother("salt", 1) is False

        mock = MagicMock(return_value={"other": "foobar"})
        with patch.object(useradd, "_get_gecos", mock):
            assert useradd.chother("salt", "foobar") is True

        mock = MagicMock(return_value={"other": "foobar2"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"other": "foobar3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chother("salt", "foobar") is False

        mock = MagicMock(return_value={"other": "foobar3"})
        with patch.object(useradd, "_get_gecos", mock):
            mock = MagicMock(return_value=None)
            with patch.dict(useradd.__salt__, {"cmd.run": mock}):
                mock = MagicMock(return_value={"other": "foobar3"})
                with patch.object(useradd, "info", mock):
                    assert useradd.chother("salt", "foobar") is False



"""
Support for Tomcat

This module uses the manager webapp to manage Apache tomcat webapps.
If the manager webapp is not configured some of the functions won't work.

:configuration:
    - Java bin path should be in default path
    - If ipv6 is enabled make sure you permit manager access to ipv6 interface
      "0:0:0:0:0:0:0:1"
    - If you are using tomcat.tar.gz it has to be installed or symlinked under
      ``/opt``, preferably using name tomcat
    - "tomcat.signal start/stop" works but it does not use the startup scripts

The following grains/pillar should be set:

.. code-block:: yaml

    tomcat-manager:
      user: <username>
      passwd: <password>

or the old format:

.. code-block:: yaml

    tomcat-manager.user: <username>
    tomcat-manager.passwd: <password>

Also configure a user in the conf/tomcat-users.xml file:

.. code-block:: xml

    <?xml version='1.0' encoding='utf-8'?>
    <tomcat-users>
        <role rolename="manager-script"/>
        <user username="tomcat" password="tomcat" roles="manager-script"/>
    </tomcat-users>

.. note::

   - More information about tomcat manager:
     http://tomcat.apache.org/tomcat-7.0-doc/manager-howto.html
   - if you use only this module for deployments you've might want to strict
     access to the manager only from localhost for more info:
     http://tomcat.apache.org/tomcat-7.0-doc/manager-howto.html#Configuring_Manager_Application_Access
   - Tested on:

     JVM Vendor:
         Sun Microsystems Inc.
     JVM Version:
         1.6.0_43-b01
     OS Architecture:
         amd64
     OS Name:
         Linux
     OS Version:
         2.6.32-358.el6.x86_64
     Tomcat Version:
         Apache Tomcat/7.0.37
"""

import glob
import hashlib
import logging
import os
import re
import tempfile
import urllib.parse
import urllib.request

import salt.utils.data
import salt.utils.stringutils

log = logging.getLogger(__name__)

__func_alias__ = {"reload_": "reload"}

# Support old-style grains/pillar
# config as well as new.
__valid_configs = {
    "user": ["tomcat-manager.user", "tomcat-manager:user"],
    "passwd": ["tomcat-manager.passwd", "tomcat-manager:passwd"],
}


def __virtual__():
    """
    Only load tomcat if it is installed or if grains/pillar config exists
    """
    if __catalina_home() or _auth("dummy"):
        return "tomcat"
    return (
        False,
        "Tomcat execution module not loaded: neither Tomcat installed locally nor"
        " tomcat-manager credentials set in grains/pillar/config.",
    )


def __catalina_home():
    """
    Tomcat paths differ depending on packaging
    """
    locations = ["/usr/share/tomcat*", "/opt/tomcat"]
    for location in locations:
        folders = glob.glob(location)
        if folders:
            for catalina_home in folders:
                if os.path.isdir(catalina_home + "/bin"):
                    return catalina_home
    return False


def _get_credentials():
    """
    Get the username and password from opts, grains, or pillar
    """
    ret = {"user": False, "passwd": False}

    # Loop through opts, grains, and pillar
    # Return the first acceptable configuration found
    for item in ret:
        for struct in [__opts__, __grains__, __pillar__]:
            # Look for the config key
            # Support old-style config format and new
            for config_key in __valid_configs[item]:
                value = salt.utils.data.traverse_dict_and_list(struct, config_key, None)
                if value:
                    ret[item] = value
                    break
    return ret["user"], ret["passwd"]


def _auth(uri):
    """
    returns a authentication handler.
    Get user & password from grains, if are not set default to
    modules.config.option

    If user & pass are missing return False
    """

    user, password = _get_credentials()
    if user is False or password is False:
        return False

    basic = urllib.request.HTTPBasicAuthHandler()
    basic.add_password(
        realm="Tomcat Manager Application", uri=uri, user=user, passwd=password
    )
    digest = urllib.request.HTTPDigestAuthHandler()
    digest.add_password(
        realm="Tomcat Manager Application", uri=uri, user=user, passwd=password
    )
    return urllib.request.build_opener(basic, digest)


def extract_war_version(war):
    """
    Extract the version from the war file name. There does not seem to be a
    standard for encoding the version into the `war file name`_

    .. _`war file name`: https://tomcat.apache.org/tomcat-6.0-doc/deployer-howto.html

    Examples:

    .. code-block:: bash

        /path/salt-2015.8.6.war -> 2015.8.6
        /path/V6R2013xD5.war -> None
    """
    basename = os.path.basename(war)
    war_package = os.path.splitext(basename)[0]  # remove '.war'
    version = re.findall("-([\\d.-]+)$", war_package)  # try semver
    return version[0] if version and len(version) == 1 else None  # default to none


def _wget(cmd, opts=None, url="http://localhost:8080/manager", timeout=180):
    """
    A private function used to issue the command to tomcat via the manager
    webapp

    cmd
        the command to execute

    url
        The URL of the server manager webapp (example:
        http://localhost:8080/manager)

    opts
        a dict of arguments

    timeout
        timeout for HTTP request

    Return value is a dict in the from of::

        {
            res: [True|False]
            msg: list of lines we got back from the manager
        }
    """

    ret = {"res": True, "msg": []}

    # prepare authentication
    auth = _auth(url)
    if auth is False:
        ret["res"] = False
        ret["msg"] = "missing username and password settings (grain/pillar)"
        return ret

    # prepare URL
    if url[-1] != "/":
        url += "/"
    url6 = url
    url += "text/{}".format(cmd)
    url6 += "{}".format(cmd)
    if opts:
        url += "?{}".format(urllib.parse.urlencode(opts))
        url6 += "?{}".format(urllib.parse.urlencode(opts))

    # Make the HTTP request
    urllib.request.install_opener(auth)

    try:
        # Trying tomcat >= 7 url
        ret["msg"] = urllib.request.urlopen(url, timeout=timeout).read().splitlines()
    except Exception:  # pylint: disable=broad-except
        try:
            # Trying tomcat6 url
            ret["msg"] = (
                urllib.request.urlopen(url6, timeout=timeout).read().splitlines()
            )
        except Exception:  # pylint: disable=broad-except
            ret["msg"] = "Failed to create HTTP request"

    # Force all byte strings to utf-8 strings, for python >= 3.4
    for key, value in enumerate(ret["msg"]):
        try:
            ret["msg"][key] = salt.utils.stringutils.to_unicode(value, "utf-8")
        except (UnicodeDecodeError, AttributeError):
            pass

    if not ret["msg"][0].startswith("OK"):
        ret["res"] = False

    return ret


def _simple_cmd(cmd, app, url="http://localhost:8080/manager", timeout=180):
    """
    Simple command wrapper to commands that need only a path option
    """

    try:
        opts = {"path": app, "version": ls(url)[app]["version"]}
        return "\n".join(_wget(cmd, opts, url, timeout=timeout)["msg"])
    except Exception:  # pylint: disable=broad-except
        return "FAIL - No context exists for path {}".format(app)


# Functions
def leaks(url="http://localhost:8080/manager", timeout=180):
    """
    Find memory leaks in tomcat

    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.leaks
    """

    return _wget("findleaks", {"statusLine": "true"}, url, timeout=timeout)["msg"]


def status(url="http://localhost:8080/manager", timeout=180):
    """
    Used to test if the tomcat manager is up

    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.status
        salt '*' tomcat.status http://localhost:8080/manager
    """

    return _wget("list", {}, url, timeout=timeout)["res"]


def ls(url="http://localhost:8080/manager", timeout=180):
    """
    list all the deployed webapps

    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.ls
        salt '*' tomcat.ls http://localhost:8080/manager
    """

    ret = {}
    data = _wget("list", "", url, timeout=timeout)
    if data["res"] is False:
        return {}
    data["msg"].pop(0)
    for line in data["msg"]:
        tmp = line.split(":")
        ret[tmp[0]] = {
            "mode": tmp[1],
            "sessions": tmp[2],
            "fullname": tmp[3],
            "version": "",
        }
        sliced = tmp[3].split("##")
        if len(sliced) > 1:
            ret[tmp[0]]["version"] = sliced[1]

    return ret


def stop(app, url="http://localhost:8080/manager", timeout=180):
    """
    Stop the webapp

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.stop /jenkins
        salt '*' tomcat.stop /jenkins http://localhost:8080/manager
    """

    return _simple_cmd("stop", app, url, timeout=timeout)


def start(app, url="http://localhost:8080/manager", timeout=180):
    """
    Start the webapp

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.start /jenkins
        salt '*' tomcat.start /jenkins http://localhost:8080/manager
    """

    return _simple_cmd("start", app, url, timeout=timeout)


def reload_(app, url="http://localhost:8080/manager", timeout=180):
    """
    Reload the webapp

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.reload /jenkins
        salt '*' tomcat.reload /jenkins http://localhost:8080/manager
    """

    return _simple_cmd("reload", app, url, timeout=timeout)


def sessions(app, url="http://localhost:8080/manager", timeout=180):
    """
    return the status of the webapp sessions

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.sessions /jenkins
        salt '*' tomcat.sessions /jenkins http://localhost:8080/manager
    """

    return _simple_cmd("sessions", app, url, timeout=timeout)


def status_webapp(app, url="http://localhost:8080/manager", timeout=180):
    """
    return the status of the webapp (stopped | running | missing)

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.status_webapp /jenkins
        salt '*' tomcat.status_webapp /jenkins http://localhost:8080/manager
    """

    webapps = ls(url, timeout=timeout)
    for i in webapps:
        if i == app:
            return webapps[i]["mode"]

    return "missing"


def serverinfo(url="http://localhost:8080/manager", timeout=180):
    """
    return details about the server

    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.serverinfo
        salt '*' tomcat.serverinfo http://localhost:8080/manager
    """

    data = _wget("serverinfo", {}, url, timeout=timeout)
    if data["res"] is False:
        return {"error": data["msg"]}

    ret = {}
    data["msg"].pop(0)
    for line in data["msg"]:
        tmp = line.split(":")
        ret[tmp[0].strip()] = tmp[1].strip()

    return ret


def undeploy(app, url="http://localhost:8080/manager", timeout=180):
    """
    Undeploy a webapp

    app
        the webapp context path
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    timeout : 180
        timeout for HTTP request

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.undeploy /jenkins
        salt '*' tomcat.undeploy /jenkins http://localhost:8080/manager
    """

    return _simple_cmd("undeploy", app, url, timeout=timeout)


def deploy_war(
    war,
    context,
    force="no",
    url="http://localhost:8080/manager",
    saltenv="base",
    timeout=180,
    temp_war_location=None,
    version=True,
):
    """
    Deploy a WAR file

    war
        absolute path to WAR file (should be accessible by the user running
        tomcat) or a path supported by the salt.modules.cp.get_file function
    context
        the context path to deploy
    force : False
        set True to deploy the webapp even one is deployed in the context
    url : http://localhost:8080/manager
        the URL of the server manager webapp
    saltenv : base
        the environment for WAR file in used by salt.modules.cp.get_url
        function
    timeout : 180
        timeout for HTTP request
    temp_war_location : None
        use another location to temporarily copy to war file
        by default the system's temp directory is used
    version : ''
        Specify the war version.  If this argument is provided, it overrides
        the version encoded in the war file name, if one is present.

        Examples:

        .. code-block:: bash

            salt '*' tomcat.deploy_war salt://salt-2015.8.6.war version=2015.08.r6

        .. versionadded:: 2015.8.6

    CLI Examples:

    cp module

    .. code-block:: bash

        salt '*' tomcat.deploy_war salt://application.war /api
        salt '*' tomcat.deploy_war salt://application.war /api no
        salt '*' tomcat.deploy_war salt://application.war /api yes http://localhost:8080/manager

    minion local file system

    .. code-block:: bash

        salt '*' tomcat.deploy_war /tmp/application.war /api
        salt '*' tomcat.deploy_war /tmp/application.war /api no
        salt '*' tomcat.deploy_war /tmp/application.war /api yes http://localhost:8080/manager
    """
    # Decide the location to copy the war for the deployment
    tfile = "salt.{}".format(os.path.basename(war))
    if temp_war_location is not None:
        if not os.path.isdir(temp_war_location):
            return 'Error - "{}" is not a directory'.format(temp_war_location)
        tfile = os.path.join(temp_war_location, tfile)
    else:
        tfile = os.path.join(tempfile.gettempdir(), tfile)

    # Copy file name if needed
    cache = False
    if not os.path.isfile(war):
        cache = True
        cached = __salt__["cp.get_url"](war, tfile, saltenv)
        if not cached:
            return "FAIL - could not cache the WAR file"
        try:
            __salt__["file.set_mode"](cached, "0644")
        except KeyError:
            pass
    else:
        tfile = war

    # Prepare options
    opts = {
        "war": "file:{}".format(tfile),
        "path": context,
    }

    # If parallel versions are desired or not disabled
    if version:
        # Set it to defined version or attempt extract
        version = extract_war_version(war) if version is True else version

        if isinstance(version, str):
            # Only pass version to Tomcat if not undefined
            opts["version"] = version

    if force == "yes":
        opts["update"] = "true"

    # Deploy
    deployed = _wget("deploy", opts, url, timeout=timeout)
    res = "\n".join(deployed["msg"])

    # Cleanup
    if cache:
        __salt__["file.remove"](tfile)

    return res


def passwd(passwd, user="", alg="sha1", realm=None):
    """
    This function replaces the $CATALINA_HOME/bin/digest.sh script
    convert a clear-text password to the $CATALINA_BASE/conf/tomcat-users.xml
    format

    CLI Examples:

    .. code-block:: bash

        salt '*' tomcat.passwd secret
        salt '*' tomcat.passwd secret tomcat sha1
        salt '*' tomcat.passwd secret tomcat sha1 'Protected Realm'
    """
    # Shouldn't it be SHA265 instead of SHA1?
    digest = hasattr(hashlib, alg) and getattr(hashlib, alg) or None
    if digest:
        if realm:
            digest.update(
                "{}:{}:{}".format(
                    user,
                    realm,
                    passwd,
                )
            )
        else:
            digest.update(passwd)

    return digest and digest.hexdigest() or False


# Non-Manager functions
def version():
    """
    Return server version from catalina.sh version

    CLI Example:

    .. code-block:: bash

        salt '*' tomcat.version
    """
    cmd = __catalina_home() + "/bin/catalina.sh version"
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        if not line:
            continue
        if "Server version" in line:
            comps = line.split(": ")
            return comps[1]


def fullversion():
    """
    Return all server information from catalina.sh version

    CLI Example:

    .. code-block:: bash

        salt '*' tomcat.fullversion
    """
    cmd = __catalina_home() + "/bin/catalina.sh version"
    ret = {}
    out = __salt__["cmd.run"](cmd).splitlines()
    for line in out:
        if not line:
            continue
        if ": " in line:
            comps = line.split(": ")
            ret[comps[0]] = comps[1].lstrip()
    return ret


def signal(signal=None):
    """
    Signals catalina to start, stop, securestart, forcestop.

    CLI Example:

    .. code-block:: bash

        salt '*' tomcat.signal start
    """
    valid_signals = {
        "forcestop": "stop -force",
        "securestart": "start -security",
        "start": "start",
        "stop": "stop",
    }

    if signal not in valid_signals:
        return

    cmd = "{}/bin/catalina.sh {}".format(__catalina_home(), valid_signals[signal])
    __salt__["cmd.run"](cmd)


if __name__ == "__main__":
    # Allow testing from the CLI
    __opts__ = {}
    __grains__ = {}
    __pillar__ = {
        "tomcat-manager.user": "foobar",
        "tomcat-manager.passwd": "barfoo1!",
    }

    old_format_creds = _get_credentials()

    __pillar__ = {"tomcat-manager": {"user": "foobar", "passwd": "barfoo1!"}}

    new_format_creds = _get_credentials()

    if old_format_creds == new_format_creds:
        log.info("Config backwards compatible")
    else:
        log.ifno("Config not backwards compatible")



"""
Test case for the cassandra_cql module
"""


import logging
import ssl

import pytest

import salt.modules.cassandra_cql as cassandra_cql
from salt.exceptions import CommandExecutionError
from tests.support.mock import MagicMock, patch

log = logging.getLogger(__name__)


pytestmark = pytest.mark.skipif(
    not cassandra_cql.HAS_DRIVER, reason="Cassandra CQL driver not loaded"
)


@pytest.fixture
def configure_loader_modules():
    return {cassandra_cql: {}}


def test_cql_query(caplog):
    """
    Test salt.modules.cassandra_cql.cql_query function
    """

    mock_session = MagicMock()
    mock_client = MagicMock()
    mock = MagicMock(return_value=(mock_session, mock_client))
    query = "query"
    with patch.object(cassandra_cql, "_connect", mock):
        query_result = cassandra_cql.cql_query(query)

    assert query_result == []

    query = {"5.0.1": "query1", "5.0.0": "query2"}
    mock_version = MagicMock(return_value="5.0.1")
    mock_session = MagicMock()
    mock_client = MagicMock()
    mock = MagicMock(return_value=(mock_session, mock_client))
    with patch.object(cassandra_cql, "version", mock_version):
        with patch.object(cassandra_cql, "_connect", mock):
            query_result = cassandra_cql.cql_query(query)
    assert query_result == []


def test_cql_query_with_prepare(caplog):
    """
    Test salt.modules.cassandra_cql.cql_query_with_prepare function
    """

    mock_session = MagicMock()
    mock_client = MagicMock()
    mock = MagicMock(return_value=(mock_session, mock_client))
    query = "query"
    statement_args = {"arg1": "test"}

    mock_context = MagicMock(
        return_value={"cassandra_cql_prepared": {"statement_name": query}}
    )
    with patch.object(cassandra_cql, "__context__", mock_context):
        with patch.object(cassandra_cql, "_connect", mock):
            query_result = cassandra_cql.cql_query_with_prepare(
                query, "statement_name", statement_args
            )
    assert query_result == []


def test_version(caplog):
    """
    Test salt.modules.cassandra_cql.version function
    """
    mock_cql_query = MagicMock(return_value=[{"release_version": "5.0.1"}])

    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        version = cassandra_cql.version()
    assert version == "5.0.1"

    mock_cql_query = MagicMock(side_effect=CommandExecutionError)
    with pytest.raises(CommandExecutionError) as err:
        with patch.object(cassandra_cql, "cql_query", mock_cql_query):
            version = cassandra_cql.version()
    assert "{}".format(err.value) == ""
    assert "Could not get Cassandra version." in caplog.text
    for record in caplog.records:
        assert record.levelname == "CRITICAL"


def test_info():
    """
    Test salt.modules.cassandra_cql.info function
    """
    expected = {"result": "info"}
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        info = cassandra_cql.info()

    assert info == expected


def test_list_keyspaces():
    """
    Test salt.modules.cassandra_cql.list_keyspaces function
    """
    expected = [{"keyspace_name": "name1"}, {"keyspace_name": "name2"}]
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        keyspaces = cassandra_cql.list_keyspaces()

    assert keyspaces == expected


def test_list_column_families():
    """
    Test salt.modules.cassandra_cql.list_column_families function
    """
    expected = [{"colum_name": "column1"}, {"column_name": "column2"}]
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        columns = cassandra_cql.list_column_families()

    assert columns == expected


def test_keyspace_exists():
    """
    Test salt.modules.cassandra_cql.keyspace_exists function
    """
    expected = "keyspace"
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        exists = cassandra_cql.keyspace_exists("keyspace")

    assert exists == bool(expected)

    expected = []
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        exists = cassandra_cql.keyspace_exists("keyspace")

    assert exists == bool(expected)


def test_create_keyspace():
    """
    Test salt.modules.cassandra_cql.create_keyspace function
    """
    expected = None
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.create_keyspace("keyspace")

    assert result == expected


def test_drop_keyspace():
    """
    Test salt.modules.cassandra_cql.drop_keyspace function
    """
    expected = True
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.drop_keyspace("keyspace")

    assert result == expected


def test_list_users():
    """
    Test salt.modules.cassandra_cql.list_users function
    """
    expected = [{"name": "user1", "super": True}, {"name": "user2", "super": False}]
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.list_users()

    assert result == expected


def test_create_user():
    """
    Test salt.modules.cassandra_cql.create_user function
    """
    expected = True
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.create_user("user", "password")

    assert result == expected


def test_list_permissions():
    """
    Test salt.modules.cassandra_cql.list_permissions function
    """
    expected = [
        {
            "permission": "ALTER",
            "resource": "<keyspace one>",
            "role": "user1",
            "username": "user1",
        }
    ]
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.list_permissions(username="user1", resource="one")

    assert result == expected


def test_grant_permission():
    """
    Test salt.modules.cassandra_cql.grant_permission function
    """
    expected = True
    mock_cql_query = MagicMock(return_value=expected)
    with patch.object(cassandra_cql, "cql_query", mock_cql_query):
        result = cassandra_cql.grant_permission(username="user1", resource="one")

    assert result == expected


def test_returns_opts_if_specified():
    """
    If ssl options are present then check that they are parsed and returned
    """
    options = MagicMock(
        return_value={
            "cluster": ["192.168.50.10", "192.168.50.11", "192.168.50.12"],
            "port": 9000,
            "ssl_options": {
                "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
                "ssl_version": "PROTOCOL_TLSv1",
            },
            "username": "cas_admin",
        }
    )

    with patch.dict(cassandra_cql.__salt__, {"config.option": options}):

        assert cassandra_cql._get_ssl_opts() == {  # pylint: disable=protected-access
            "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
            "ssl_version": ssl.PROTOCOL_TLSv1,
        }  # pylint: disable=no-member


def test_invalid_protocol_version():
    """
    Check that the protocol version is imported only if it isvalid
    """
    options = MagicMock(
        return_value={
            "cluster": ["192.168.50.10", "192.168.50.11", "192.168.50.12"],
            "port": 9000,
            "ssl_options": {
                "ca_certs": "/etc/ssl/certs/ca-bundle.trust.crt",
                "ssl_version": "Invalid",
            },
            "username": "cas_admin",
        }
    )

    with patch.dict(cassandra_cql.__salt__, {"config.option": options}):
        with pytest.raises(CommandExecutionError):
            cassandra_cql._get_ssl_opts()  # pylint: disable=protected-access


def test_unspecified_opts():
    """
    Check that it returns None when ssl opts aren't specified
    """
    with patch.dict(
        cassandra_cql.__salt__, {"config.option": MagicMock(return_value={})}
    ):
        assert cassandra_cql._get_ssl_opts() is None


def test_valid_asynchronous_args():
    mock_execute = MagicMock(return_value={})
    mock_execute_async = MagicMock(return_value={})
    mock_context = {
        "cassandra_cql_returner_cluster": MagicMock(return_value={}),
        "cassandra_cql_returner_session": MagicMock(
            execute=mock_execute,
            execute_async=mock_execute_async,
            prepare=lambda _: MagicMock(bind=lambda _: None),  # mock prepared_statement
            row_factory=None,
        ),
        "cassandra_cql_prepared": {},
    }

    with patch.dict(cassandra_cql.__context__, mock_context):
        cassandra_cql.cql_query_with_prepare(
            "SELECT now() from system.local;", "select_now", [], asynchronous=True
        )
        mock_execute_async.assert_called_once()


def test_valid_async_args():
    mock_execute = MagicMock(return_value={})
    mock_execute_async = MagicMock(return_value={})
    mock_context = {
        "cassandra_cql_returner_cluster": MagicMock(return_value={}),
        "cassandra_cql_returner_session": MagicMock(
            execute=mock_execute,
            execute_async=mock_execute_async,
            prepare=lambda _: MagicMock(bind=lambda _: None),
            # mock prepared_statement
            row_factory=None,
        ),
        "cassandra_cql_prepared": {},
    }

    kwargs = {"async": True}  # to avoid syntax error in python 3.7
    with patch.dict(cassandra_cql.__context__, mock_context):
        cassandra_cql.cql_query_with_prepare(
            "SELECT now() from system.local;", "select_now", [], **kwargs
        )
        mock_execute_async.assert_called_once()



"""
Test the lxc module
"""

import pytest

from tests.support.case import ModuleCase


@pytest.mark.skip(
    reason="Needs rewrite to be more distro agnostic. Also, the tearDown "
    "function destroys ALL containers on the box, which is BAD.",
)
@pytest.mark.windows_whitelisted
@pytest.mark.skip_if_not_root
@pytest.mark.skip_if_binaries_missing(
    "lxc-start", reason="LXC is not installed or minimal version not met"
)
class LXCModuleTest(ModuleCase):
    """
    Test the lxc module
    """

    prefix = "_salttesting"

    def setUp(self):
        os = self.run_function("grains.item", ["os", "oscodename", "osarch"])

        p = {
            "download": {
                "dist": os["os"].lower(),
                "arch": os["osarch"].lower(),
                "template": "download",
                "release": os["oscodename"].lower(),
            },
            "sshd": {"template": "sshd"},
        }
        self.run_function("grains.setval", ["lxc.profile", p])

    def tearDown(self):
        """
        Clean up any LXCs created.
        """
        r = self.run_function("lxc.list")
        for k, v in r.items():
            for x in v:
                if x.startswith(self.prefix):
                    self.run_function("lxc.destroy", [x])

    def test_create_destroy(self):
        """
        Test basic create/destroy of an LXC.
        """

        r = self.run_function("lxc.create", [self.prefix], template="sshd")
        self.assertEqual(r, {"state": {"new": "stopped", "old": None}, "result": True})
        self.assertTrue(self.run_function("lxc.exists", [self.prefix]))
        r = self.run_function("lxc.destroy", [self.prefix])
        self.assertEqual(r, {"state": None, "change": True})
        self.assertFalse(self.run_function("lxc.exists", [self.prefix]))

    def test_init(self):
        """
        Test basic init functionality.
        """

        r = self.run_function("lxc.init", [self.prefix], profile="sshd", seed=False)
        self.assertTrue(r.get("created", False))
        self.assertTrue(self.run_function("lxc.exists", [self.prefix]))

    def test_macvlan(self):
        """
        Regression test for macvlan nic profile.
        """

        p = {
            "macvlan": {
                "eth0": {"macvlan.mode": "bridge", "link": "eth0", "type": "macvlan"}
            }
        }

        self.run_function("grains.setval", ["lxc.nic", p])

        self.run_function(
            "lxc.init",
            [self.prefix],
            profile="sshd",
            nic="macvlan",
            seed=False,
            start=False,
        )

        f = "/var/lib/lxc/{}/config".format(self.prefix)
        conf = self.run_function("lxc.read_conf", [f])

        # Due to a segfault in lxc-destroy caused by invalid configs,
        # truncate the config.
        self.run_function("cmd.run", ["truncate -s 0 {}".format(f)])

        self.assertEqual(conf.get("lxc.network.type"), "macvlan")



"""
Return config information
"""

import copy
import fnmatch
import logging
import os
import urllib.parse

import salt.config
import salt.syspaths as syspaths
import salt.utils.data
import salt.utils.dictupdate
import salt.utils.files
import salt.utils.platform
import salt.utils.sdb as sdb
from salt.loader.context import LoaderContext

try:
    # Gated for salt-ssh (salt.utils.cloud imports msgpack)
    import salt.utils.cloud

    HAS_CLOUD = True
except ImportError:
    HAS_CLOUD = False


if salt.utils.platform.is_windows():
    _HOSTS_FILE = os.path.join(
        os.environ["SystemRoot"], "System32", "drivers", "etc", "hosts"
    )
else:
    _HOSTS_FILE = os.path.join(os.sep, "etc", "hosts")

log = logging.getLogger(__name__)

__proxyenabled__ = ["*"]
__salt_loader__ = LoaderContext()
__opts__ = __salt_loader__.named_context("__opts__")

# Set up the default values for all systems
DEFAULTS = {
    "mongo.db": "salt",
    "mongo.password": "",
    "mongo.port": 27017,
    "mongo.user": "",
    "redis.db": "0",
    "redis.host": "salt",
    "redis.port": 6379,
    "test.foo": "unconfigured",
    "ca.cert_base_path": "/etc/pki",
    "solr.cores": [],
    "solr.host": "localhost",
    "solr.port": "8983",
    "solr.baseurl": "/solr",
    "solr.type": "master",
    "solr.request_timeout": None,
    "solr.init_script": "/etc/rc.d/solr",
    "solr.dih.import_options": {
        "clean": False,
        "optimize": True,
        "commit": True,
        "verbose": False,
    },
    "solr.backup_path": None,
    "solr.num_backups": 1,
    "poudriere.config": "/usr/local/etc/poudriere.conf",
    "poudriere.config_dir": "/usr/local/etc/poudriere.d",
    "ldap.uri": "",
    "ldap.server": "localhost",
    "ldap.port": "389",
    "ldap.tls": False,
    "ldap.no_verify": False,
    "ldap.anonymous": True,
    "ldap.scope": 2,
    "ldap.attrs": None,
    "ldap.binddn": "",
    "ldap.bindpw": "",
    "hosts.file": _HOSTS_FILE,
    "aliases.file": "/etc/aliases",
    "virt": {
        "tunnel": False,
        "images": os.path.join(syspaths.SRV_ROOT_DIR, "salt-images"),
    },
    "docker.exec_driver": "docker-exec",
    "docker.compare_container_networks": {
        "static": ["Aliases", "Links", "IPAMConfig"],
        "automatic": ["IPAddress", "Gateway", "GlobalIPv6Address", "IPv6Gateway"],
    },
}


def backup_mode(backup=""):
    """
    Return the backup mode

    CLI Example:

    .. code-block:: bash

        salt '*' config.backup_mode
    """
    if backup:
        return backup
    return option("backup_mode")


def manage_mode(mode):
    """
    Return a mode value, normalized to a string

    CLI Example:

    .. code-block:: bash

        salt '*' config.manage_mode
    """
    # config.manage_mode should no longer be invoked from the __salt__ dunder
    # in Salt code, this function is only being left here for backwards
    # compatibility.
    return salt.utils.files.normalize_mode(mode)


def valid_fileproto(uri):
    """
    Returns a boolean value based on whether or not the URI passed has a valid
    remote file protocol designation

    CLI Example:

    .. code-block:: bash

        salt '*' config.valid_fileproto salt://path/to/file
    """
    return urllib.parse.urlparse(uri).scheme in salt.utils.files.VALID_PROTOS


def option(
    value,
    default=None,
    omit_opts=False,
    omit_grains=False,
    omit_pillar=False,
    omit_master=False,
    omit_all=False,
    wildcard=False,
):
    """
    Returns the setting for the specified config value. The priority for
    matches is the same as in :py:func:`config.get <salt.modules.config.get>`,
    only this function does not recurse into nested data structures. Another
    difference between this function and :py:func:`config.get
    <salt.modules.config.get>` is that it comes with a set of "sane defaults".
    To view these, you can run the following command:

    .. code-block:: bash

        salt '*' config.option '*' omit_all=True wildcard=True

    default
        The default value if no match is found. If not specified, then the
        fallback default will be an empty string, unless ``wildcard=True``, in
        which case the return will be an empty dictionary.

    omit_opts : False
        Pass as ``True`` to exclude matches from the minion configuration file

    omit_grains : False
        Pass as ``True`` to exclude matches from the grains

    omit_pillar : False
        Pass as ``True`` to exclude matches from the pillar data

    omit_master : False
        Pass as ``True`` to exclude matches from the master configuration file

    omit_all : True
        Shorthand to omit all of the above and return matches only from the
        "sane defaults".

        .. versionadded:: 3000

    wildcard : False
        If used, this will perform pattern matching on keys. Note that this
        will also significantly change the return data. Instead of only a value
        being returned, a dictionary mapping the matched keys to their values
        is returned. For example, using ``wildcard=True`` with a ``key`` of
        ``'foo.ba*`` could return a dictionary like so:

        .. code-block:: python

            {'foo.bar': True, 'foo.baz': False}

        .. versionadded:: 3000

    CLI Example:

    .. code-block:: bash

        salt '*' config.option redis.host
    """
    if omit_all:
        omit_opts = omit_grains = omit_pillar = omit_master = True

    if default is None:
        default = "" if not wildcard else {}

    if not wildcard:
        if not omit_opts:
            if value in __opts__:
                return __opts__[value]
        if not omit_grains:
            if value in __grains__:
                return __grains__[value]
        if not omit_pillar:
            if value in __pillar__:
                return __pillar__[value]
        if not omit_master:
            if value in __pillar__.get("master", {}):
                return __pillar__["master"][value]
        if value in DEFAULTS:
            return DEFAULTS[value]

        # No match
        return default
    else:
        # We need to do the checks in the reverse order so that minion opts
        # takes precedence
        ret = {}
        for omit, data in (
            (omit_master, __pillar__.get("master", {})),
            (omit_pillar, __pillar__),
            (omit_grains, __grains__),
            (omit_opts, __opts__),
        ):
            if not omit:
                ret.update({x: data[x] for x in fnmatch.filter(data, value)})
        # Check the DEFAULTS as well to see if the pattern matches it
        for item in (x for x in fnmatch.filter(DEFAULTS, value) if x not in ret):
            ret[item] = DEFAULTS[item]

        # If no matches, return the default
        return ret or default


def merge(value, default="", omit_opts=False, omit_master=False, omit_pillar=False):
    """
    Retrieves an option based on key, merging all matches.

    Same as ``option()`` except that it merges all matches, rather than taking
    the first match.

    CLI Example:

    .. code-block:: bash

        salt '*' config.merge schedule
    """
    ret = None
    if not omit_opts:
        if value in __opts__:
            ret = __opts__[value]
            if isinstance(ret, str):
                return ret
    if not omit_master:
        if value in __pillar__.get("master", {}):
            tmp = __pillar__["master"][value]
            if ret is None:
                ret = tmp
                if isinstance(ret, str):
                    return ret
            elif isinstance(ret, dict) and isinstance(tmp, dict):
                tmp.update(ret)
                ret = tmp
            elif isinstance(ret, (list, tuple)) and isinstance(tmp, (list, tuple)):
                ret = list(ret) + list(tmp)
    if not omit_pillar:
        if value in __pillar__:
            tmp = __pillar__[value]
            if ret is None:
                ret = tmp
                if isinstance(ret, str):
                    return ret
            elif isinstance(ret, dict) and isinstance(tmp, dict):
                tmp.update(ret)
                ret = tmp
            elif isinstance(ret, (list, tuple)) and isinstance(tmp, (list, tuple)):
                ret = list(ret) + list(tmp)
    if ret is None and value in DEFAULTS:
        return DEFAULTS[value]
    if ret is None:
        return default
    return ret


def get(
    key,
    default="",
    delimiter=":",
    merge=None,
    omit_opts=False,
    omit_pillar=False,
    omit_master=False,
    omit_grains=False,
):
    """
    .. versionadded:: 0.14.0

    Attempt to retrieve the named value from the minion config file, pillar,
    grains or the master config. If the named value is not available, return
    the value specified by the ``default`` argument. If this argument is not
    specified, ``default`` falls back to an empty string.

    Values can also be retrieved from nested dictionaries. Assume the below
    data structure:

    .. code-block:: python

        {'pkg': {'apache': 'httpd'}}

    To retrieve the value associated with the ``apache`` key, in the
    sub-dictionary corresponding to the ``pkg`` key, the following command can
    be used:

    .. code-block:: bash

        salt myminion config.get pkg:apache

    The ``:`` (colon) is used to represent a nested dictionary level.

    .. versionchanged:: 2015.5.0
        The ``delimiter`` argument was added, to allow delimiters other than
        ``:`` to be used.

    This function traverses these data stores in this order, returning the
    first match found:

    - Minion configuration
    - Minion's grains
    - Minion's pillar data
    - Master configuration (requires :conf_minion:`pillar_opts` to be set to
      ``True`` in Minion config file in order to work)

    This means that if there is a value that is going to be the same for the
    majority of minions, it can be configured in the Master config file, and
    then overridden using the grains, pillar, or Minion config file.

    Adding config options to the Master or Minion configuration file is easy:

    .. code-block:: yaml

        my-config-option: value
        cafe-menu:
          - egg and bacon
          - egg sausage and bacon
          - egg and spam
          - egg bacon and spam
          - egg bacon sausage and spam
          - spam bacon sausage and spam
          - spam egg spam spam bacon and spam
          - spam sausage spam spam bacon spam tomato and spam

    .. note::
        Minion configuration options built into Salt (like those defined
        :ref:`here <configuration-salt-minion>`) will *always* be defined in
        the Minion configuration and thus *cannot be overridden by grains or
        pillar data*. However, additional (user-defined) configuration options
        (as in the above example) will not be in the Minion configuration by
        default and thus can be overridden using grains/pillar data by leaving
        the option out of the minion config file.

    **Arguments**

    delimiter
        .. versionadded:: 2015.5.0

        Override the delimiter used to separate nested levels of a data
        structure.

    merge
        .. versionadded:: 2015.5.0

        If passed, this parameter will change the behavior of the function so
        that, instead of traversing each data store above in order and
        returning the first match, the data stores are first merged together
        and then searched. The pillar data is merged into the master config
        data, then the grains are merged, followed by the Minion config data.
        The resulting data structure is then searched for a match. This allows
        for configurations to be more flexible.

        .. note::

            The merging described above does not mean that grain data will end
            up in the Minion's pillar data, or pillar data will end up in the
            master config data, etc. The data is just combined for the purposes
            of searching an amalgam of the different data stores.

        The supported merge strategies are as follows:

        - **recurse** - If a key exists in both dictionaries, and the new value
          is not a dictionary, it is replaced. Otherwise, the sub-dictionaries
          are merged together into a single dictionary, recursively on down,
          following the same criteria. For example:

          .. code-block:: python

              >>> dict1 = {'foo': {'bar': 1, 'qux': True},
                           'hosts': ['a', 'b', 'c'],
                           'only_x': None}
              >>> dict2 = {'foo': {'baz': 2, 'qux': False},
                           'hosts': ['d', 'e', 'f'],
                           'only_y': None}
              >>> merged
              {'foo': {'bar': 1, 'baz': 2, 'qux': False},
               'hosts': ['d', 'e', 'f'],
               'only_dict1': None,
               'only_dict2': None}

        - **overwrite** - If a key exists in the top level of both
          dictionaries, the new value completely overwrites the old. For
          example:

          .. code-block:: python

              >>> dict1 = {'foo': {'bar': 1, 'qux': True},
                           'hosts': ['a', 'b', 'c'],
                           'only_x': None}
              >>> dict2 = {'foo': {'baz': 2, 'qux': False},
                           'hosts': ['d', 'e', 'f'],
                           'only_y': None}
              >>> merged
              {'foo': {'baz': 2, 'qux': False},
               'hosts': ['d', 'e', 'f'],
               'only_dict1': None,
               'only_dict2': None}

    CLI Example:

    .. code-block:: bash

        salt '*' config.get pkg:apache
        salt '*' config.get lxc.container_profile:centos merge=recurse
    """
    if merge is None:
        if not omit_opts:
            ret = salt.utils.data.traverse_dict_and_list(
                __opts__, key, "_|-", delimiter=delimiter
            )
            if ret != "_|-":
                return sdb.sdb_get(ret, __opts__)

        if not omit_grains:
            ret = salt.utils.data.traverse_dict_and_list(
                __grains__, key, "_|-", delimiter
            )
            if ret != "_|-":
                return sdb.sdb_get(ret, __opts__)

        if not omit_pillar:
            ret = salt.utils.data.traverse_dict_and_list(
                __pillar__, key, "_|-", delimiter=delimiter
            )
            if ret != "_|-":
                return sdb.sdb_get(ret, __opts__)

        if not omit_master:
            ret = salt.utils.data.traverse_dict_and_list(
                __pillar__.get("master", {}), key, "_|-", delimiter=delimiter
            )
            if ret != "_|-":
                return sdb.sdb_get(ret, __opts__)

        ret = salt.utils.data.traverse_dict_and_list(
            DEFAULTS, key, "_|-", delimiter=delimiter
        )
        if ret != "_|-":
            return sdb.sdb_get(ret, __opts__)
    else:
        if merge not in ("recurse", "overwrite"):
            log.warning(
                "Unsupported merge strategy '%s'. Falling back to 'recurse'.", merge
            )
            merge = "recurse"

        merge_lists = salt.config.master_config("/etc/salt/master").get(
            "pillar_merge_lists"
        )

        data = copy.copy(DEFAULTS)
        data = salt.utils.dictupdate.merge(
            data, __pillar__.get("master", {}), strategy=merge, merge_lists=merge_lists
        )
        data = salt.utils.dictupdate.merge(
            data, __pillar__, strategy=merge, merge_lists=merge_lists
        )
        data = salt.utils.dictupdate.merge(
            data, __grains__, strategy=merge, merge_lists=merge_lists
        )
        data = salt.utils.dictupdate.merge(
            data, __opts__, strategy=merge, merge_lists=merge_lists
        )
        ret = salt.utils.data.traverse_dict_and_list(
            data, key, "_|-", delimiter=delimiter
        )
        if ret != "_|-":
            return sdb.sdb_get(ret, __opts__)

    return default


def dot_vals(value):
    """
    Pass in a configuration value that should be preceded by the module name
    and a dot, this will return a list of all read key/value pairs

    CLI Example:

    .. code-block:: bash

        salt '*' config.dot_vals host
    """
    ret = {}
    for key, val in __pillar__.get("master", {}).items():
        if key.startswith("{}.".format(value)):
            ret[key] = val
    for key, val in __opts__.items():
        if key.startswith("{}.".format(value)):
            ret[key] = val
    return ret


def gather_bootstrap_script(bootstrap=None):
    """
    Download the salt-bootstrap script, and return its location

    bootstrap
        URL of alternate bootstrap script

    CLI Example:

    .. code-block:: bash

        salt '*' config.gather_bootstrap_script
    """
    if not HAS_CLOUD:
        return False, "config.gather_bootstrap_script is unavailable"
    ret = salt.utils.cloud.update_bootstrap(__opts__, url=bootstrap)
    if "Success" in ret and ret["Success"]["Files updated"]:
        return ret["Success"]["Files updated"][0]


def items():
    """
    Return the complete config from the currently running minion process.
    This includes defaults for values not set in the config file.

    CLI Example:

    .. code-block:: bash

        salt '*' config.items
    """
    return __opts__



"""
Management of addresses and names in hosts file
===============================================

The ``/etc/hosts`` file can be managed to contain definitions for specific hosts:

.. code-block:: yaml

    salt-master:
      host.present:
        - ip: 192.168.0.42

Or using the ``names`` directive, you can put several names for the same IP.
(Do not try one name with space-separated values).

.. code-block:: yaml

    server1:
      host.present:
        - ip: 192.168.0.42
        - names:
          - server1
          - florida

.. note::

    Changing the ``names`` in ``host.present`` does not cause an
    update to remove the old entry.

.. code-block:: yaml

    server1:
      host.present:
        - ip:
          - 192.168.0.42
          - 192.168.0.43
          - 192.168.0.44
        - names:
          - server1

You can replace all existing names for a particular IP address:

.. code-block:: yaml

    127.0.1.1:
      host.only:
        - hostnames:
          - foo.example.com
          - foo

Or delete all existing names for an address:

.. code-block:: yaml

    203.0.113.25:
        host.only:
          - hostnames: []

You can also include comments:

.. code-block:: yaml

    server1:
      host.present:
        - ip: 192.168.0.42
        - names:
          - server1
          - florida
        - comment: A very important comment

"""


import logging

import salt.utils.validate.net

log = logging.getLogger(__name__)


def present(name, ip, comment="", clean=False):  # pylint: disable=C0103
    """
    Ensures that the named host is present with the given ip

    name
        The host to assign an ip to

    ip
        The ip addr(s) to apply to the host. Can be a single IP or a list of IP
        addresses.

    comment
        A comment to include for the host entry

        .. versionadded:: 3001

    clean
        Remove any entries which don't match those configured in the ``ip``
        option. Default is ``False``.

        .. versionadded:: 2018.3.4
    """
    ret = {"name": name, "changes": {}, "result": True, "comment": ""}

    if not isinstance(ip, list):
        ip = [ip]

    all_hosts = __salt__["hosts.list_hosts"]()
    comments = []
    to_add = set()
    to_remove = set()
    update_comment = set()

    # First check for IPs not currently in the hosts file
    to_add.update([(addr, name) for addr in ip if addr not in all_hosts])

    if comment:
        update_comment.update([(addr, comment) for addr in ip if addr not in all_hosts])

    # Now sweep through the hosts file and look for entries matching either the
    # IP address(es) or hostname.
    for addr, host_info in all_hosts.items():
        if addr not in ip:
            if "aliases" in host_info and name in host_info["aliases"]:
                # Found match for hostname, but the corresponding IP is not in
                # our list, so we need to remove it.
                if clean:
                    to_remove.add((addr, name))
                else:
                    ret.setdefault("warnings", []).append(
                        "Host {0} present for IP address {1}. To get rid of "
                        "this warning, either run this state with 'clean' "
                        "set to True to remove {0} from {1}, or add {1} to "
                        "the 'ip' argument.".format(name, addr)
                    )
        else:
            if "aliases" in host_info and name in host_info["aliases"]:
                if (
                    comment
                    and "comment" in host_info
                    and host_info["comment"] != comment
                ):
                    update_comment.add((addr, comment))
                elif comment and "comment" not in host_info:
                    update_comment.add((addr, comment))
                else:
                    # No changes needed for this IP address and hostname
                    comments.append("Host {} ({}) already present".format(name, addr))
            else:
                # IP address listed in hosts file, but hostname is not present.
                # We will need to add it.
                if salt.utils.validate.net.ip_addr(addr):
                    to_add.add((addr, name))
                    if comment:
                        update_comment.add((addr, comment))
                else:
                    ret["result"] = False
                    comments.append("Invalid IP Address for {} ({})".format(name, addr))

    for addr, name in to_add:
        if __opts__["test"]:
            ret["result"] = None
            comments.append("Host {} ({}) would be added".format(name, addr))
        else:
            if __salt__["hosts.add_host"](addr, name):
                comments.append("Added host {} ({})".format(name, addr))
            else:
                ret["result"] = False
                comments.append("Failed to add host {} ({})".format(name, addr))
                continue
        ret["changes"].setdefault("added", {}).setdefault(addr, []).append(name)

    for addr, comment in update_comment:
        if __opts__["test"]:
            comments.append("Comment for {} ({}) would be added".format(addr, comment))
        else:
            if __salt__["hosts.set_comment"](addr, comment):
                comments.append("Set comment for host {} ({})".format(addr, comment))
            else:
                ret["result"] = False
                comments.append(
                    "Failed to add comment for host {} ({})".format(addr, comment)
                )
                continue
        ret["changes"].setdefault("comment_added", {}).setdefault(addr, []).append(
            comment
        )

    for addr, name in to_remove:
        if __opts__["test"]:
            ret["result"] = None
            comments.append("Host {} ({}) would be removed".format(name, addr))
        else:
            if __salt__["hosts.rm_host"](addr, name):
                comments.append("Removed host {} ({})".format(name, addr))
            else:
                ret["result"] = False
                comments.append("Failed to remove host {} ({})".format(name, addr))
                continue
        ret["changes"].setdefault("removed", {}).setdefault(addr, []).append(name)

    ret["comment"] = "\n".join(comments)
    return ret


def absent(name, ip):  # pylint: disable=C0103
    """
    Ensure that the named host is absent

    name
        The host to remove

    ip
        The ip addr(s) of the host to remove
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if not isinstance(ip, list):
        ip = [ip]

    comments = []
    for _ip in ip:
        if not __salt__["hosts.has_pair"](_ip, name):
            ret["result"] = True
            comments.append("Host {} ({}) already absent".format(name, _ip))
        else:
            if __opts__["test"]:
                comments.append("Host {} ({}) needs to be removed".format(name, _ip))
            else:
                if __salt__["hosts.rm_host"](_ip, name):
                    ret["changes"] = {"host": name}
                    ret["result"] = True
                    comments.append("Removed host {} ({})".format(name, _ip))
                else:
                    ret["result"] = False
                    comments.append("Failed to remove host")
    ret["comment"] = "\n".join(comments)
    return ret


def only(name, hostnames):
    """
    Ensure that only the given hostnames are associated with the
    given IP address.

    .. versionadded:: 2016.3.0

    name
        The IP address to associate with the given hostnames.

    hostnames
        Either a single hostname or a list of hostnames to associate
        with the given IP address in the given order.  Any other
        hostname associated with the IP address is removed.  If no
        hostnames are specified, all hostnames associated with the
        given IP address are removed.
    """
    ret = {"name": name, "changes": {}, "result": None, "comment": ""}

    if isinstance(hostnames, str):
        hostnames = [hostnames]

    old = " ".join(__salt__["hosts.get_alias"](name))
    new = " ".join(x.strip() for x in hostnames)

    if old == new:
        ret["comment"] = 'IP address {} already set to "{}"'.format(name, new)
        ret["result"] = True
        return ret

    if __opts__["test"]:
        ret["comment"] = 'Would change {} from "{}" to "{}"'.format(name, old, new)
        return ret

    ret["result"] = __salt__["hosts.set_host"](name, new)
    if not ret["result"]:
        ret["comment"] = 'hosts.set_host failed to change {} from "{}" to "{}"'.format(
            name, old, new
        )
        return ret

    ret["comment"] = 'successfully changed {} from "{}" to "{}"'.format(name, old, new)
    ret["changes"] = {name: {"old": old, "new": new}}
    return ret



"""
Integration tests for DigitalOcean APIv2
"""

import base64
import hashlib

import salt.crypt
import salt.utils.stringutils
from tests.integration.cloud.helpers.cloud_test_base import TIMEOUT, CloudTest


class DigitalOceanTest(CloudTest):
    """
    Integration tests for the DigitalOcean cloud provider in Salt-Cloud
    """

    PROVIDER = "digitalocean"
    REQUIRED_PROVIDER_CONFIG_ITEMS = (
        "personal_access_token",
        "ssh_key_file",
        "ssh_key_name",
    )

    def test_list_images(self):
        """
        Tests the return of running the --list-images command for digitalocean
        """
        image_list = self.run_cloud("--list-images {}".format(self.PROVIDER))
        self.assertIn("ubuntu-18-04-x64", [i.strip() for i in image_list])

    def test_list_locations(self):
        """
        Tests the return of running the --list-locations command for digitalocean
        """
        _list_locations = self.run_cloud("--list-locations {}".format(self.PROVIDER))
        self.assertIn("San Francisco 2", [i.strip() for i in _list_locations])

    def test_list_sizes(self):
        """
        Tests the return of running the --list-sizes command for digitalocean
        """
        _list_sizes = self.run_cloud("--list-sizes {}".format(self.PROVIDER))
        self.assertIn("16gb", [i.strip() for i in _list_sizes])

    def test_key_management(self):
        """
        Test key management
        """
        do_key_name = self.instance_name + "-key"

        # generate key and fingerprint
        if salt.crypt.HAS_M2:

            rsa_key = salt.crypt.RSA.gen_key(4096, 65537, lambda: None)
            pub = salt.utils.stringutils.to_bytes(
                "ssh-rsa {}".format(
                    base64.b64encode(
                        salt.utils.stringutils.to_bytes(
                            "\x00\x00\x00\x07ssh-rsa{}{}".format(*rsa_key.pub())
                        )
                    )
                )
            )
        else:
            ssh_key = salt.crypt.RSA.generate(4096)
            pub = ssh_key.publickey().exportKey("OpenSSH")
        pub = salt.utils.stringutils.to_str(pub)
        key_hex = hashlib.md5(
            base64.b64decode(pub.strip().split()[1].encode())
        ).hexdigest()
        finger_print = ":".join([key_hex[x : x + 2] for x in range(0, len(key_hex), 2)])

        try:
            _key = self.run_cloud(
                '-f create_key {} name="{}" public_key="{}"'.format(
                    self.PROVIDER, do_key_name, pub
                )
            )

            # Upload public key
            self.assertIn(finger_print, [i.strip() for i in _key])

            # List all keys
            list_keypairs = self.run_cloud("-f list_keypairs {}".format(self.PROVIDER))

            self.assertIn(finger_print, [i.strip() for i in list_keypairs])

            # List key
            show_keypair = self.run_cloud(
                "-f show_keypair {} keyname={}".format(self.PROVIDER, do_key_name)
            )
            self.assertIn(finger_print, [i.strip() for i in show_keypair])
        except AssertionError:
            # Delete the public key if the above assertions fail
            self.run_cloud("-f remove_key {} id={}".format(self.PROVIDER, finger_print))
            raise
        finally:
            # Delete public key
            self.assertTrue(
                self.run_cloud(
                    "-f remove_key {} id={}".format(self.PROVIDER, finger_print)
                )
            )

    def test_instance(self):
        """
        Test creating an instance on DigitalOcean
        """
        # check if instance with salt installed returned
        ret_str = self.run_cloud(
            "-p digitalocean-test {}".format(self.instance_name), timeout=TIMEOUT
        )
        self.assertInstanceExists(ret_str)

        self.assertDestroyInstance()



import pathlib

import pytest
from pytestshellutils.utils.ports import get_unused_localhost_port

import salt.config
import tests.support.netapi as netapi


@pytest.fixture
def netapi_port():
    return get_unused_localhost_port()


@pytest.fixture
def client_config(salt_master_factory):
    # Make sure we have the tokens directory writable
    tokens_dir = pathlib.Path(salt_master_factory.config["cachedir"]) / "tokens"
    if not tokens_dir.is_dir():
        tokens_dir.mkdir()
    config = salt.config.client_config(
        salt_master_factory.config["conf_file"],
        defaults=salt_master_factory.config.copy(),
    )
    return config


@pytest.fixture
def minion_config(salt_minion_factory):
    return salt_minion_factory.config.copy()


@pytest.fixture
def load_auth(client_config):
    return netapi.load_auth(client_config)


@pytest.fixture(scope="package")
def salt_netapi_account(salt_netapi_account_factory):
    with salt_netapi_account_factory as account:
        yield account


@pytest.fixture
def auth_creds(salt_netapi_account):
    return {
        "username": salt_netapi_account.username,
        "password": salt_netapi_account.password,
        "eauth": "auto",
    }


@pytest.fixture
def auth_token(load_auth, auth_creds):
    """
    Mint and return a valid token for auth_creds
    """
    return netapi.auth_token(load_auth, auth_creds)


@pytest.fixture
def content_type_map():
    return netapi.content_type_map()



"""

.. warning::

    The `cassandra` module is deprecated in favor of the `cassandra_cql`
    module.

Cassandra NoSQL Database Module

:depends:   - pycassa Cassandra Python adapter
:configuration:
    The location of the 'nodetool' command, host, and thrift port needs to be
    specified via pillar::

        cassandra.nodetool: /usr/local/bin/nodetool
        cassandra.host: localhost
        cassandra.thrift_port: 9160
"""

import logging

import salt.utils.path
from salt.utils.versions import warn_until_date

log = logging.getLogger(__name__)


HAS_PYCASSA = False
try:
    from pycassa.system_manager import SystemManager

    HAS_PYCASSA = True
except ImportError:
    pass


def __virtual__():
    """
    Only load if pycassa is available and the system is configured
    """
    if not HAS_PYCASSA:
        return (
            False,
            "The cassandra execution module cannot be loaded: pycassa not installed.",
        )

    warn_until_date(
        "20240101",
        "The cassandra returner is broken and deprecated, and will be removed"
        " after {date}. Use the cassandra_cql returner instead",
    )

    if HAS_PYCASSA and salt.utils.path.which("nodetool"):
        return "cassandra"
    return (
        False,
        "The cassandra execution module cannot be loaded: nodetool not found.",
    )


def _nodetool(cmd):
    """
    Internal cassandra nodetool wrapper. Some functions are not
    available via pycassa so we must rely on nodetool.
    """
    nodetool = __salt__["config.option"]("cassandra.nodetool")
    host = __salt__["config.option"]("cassandra.host")
    return __salt__["cmd.run_stdout"]("{} -h {} {}".format(nodetool, host, cmd))


def _sys_mgr():
    """
    Return a pycassa system manager connection object
    """
    thrift_port = str(__salt__["config.option"]("cassandra.THRIFT_PORT"))
    host = __salt__["config.option"]("cassandra.host")
    return SystemManager("{}:{}".format(host, thrift_port))


def compactionstats():
    """
    Return compactionstats info

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.compactionstats
    """
    return _nodetool("compactionstats")


def version():
    """
    Return the cassandra version

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.version
    """
    return _nodetool("version")


def netstats():
    """
    Return netstats info

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.netstats
    """
    return _nodetool("netstats")


def tpstats():
    """
    Return tpstats info

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.tpstats
    """
    return _nodetool("tpstats")


def info():
    """
    Return cassandra node info

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.info
    """
    return _nodetool("info")


def ring():
    """
    Return cassandra ring info

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.ring
    """
    return _nodetool("ring")


def keyspaces():
    """
    Return existing keyspaces

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.keyspaces
    """
    sys = _sys_mgr()
    return sys.list_keyspaces()


def column_families(keyspace=None):
    """
    Return existing column families for all keyspaces
    or just the provided one.

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.column_families
        salt '*' cassandra.column_families <keyspace>
    """
    sys = _sys_mgr()
    ksps = sys.list_keyspaces()

    if keyspace:
        if keyspace in ksps:
            return list(sys.get_keyspace_column_families(keyspace).keys())
        else:
            return None
    else:
        ret = {}
        for kspace in ksps:
            ret[kspace] = list(sys.get_keyspace_column_families(kspace).keys())

        return ret


def column_family_definition(keyspace, column_family):
    """
    Return a dictionary of column family definitions for the given
    keyspace/column_family

    CLI Example:

    .. code-block:: bash

        salt '*' cassandra.column_family_definition <keyspace> <column_family>

    """
    sys = _sys_mgr()

    try:
        return vars(sys.get_keyspace_column_families(keyspace)[column_family])
    except Exception:  # pylint: disable=broad-except
        log.debug("Invalid Keyspace/CF combination")
        return None



r"""
Renderer that will decrypt GPG ciphers

Any value in the SLS file can be a GPG cipher, and this renderer will decrypt it
before passing it off to Salt. This allows you to safely store secrets in
source control, in such a way that only your Salt master can decrypt them and
distribute them only to the minions that need them.

The typical use-case would be to use ciphers in your pillar data, and keep a
secret key on your master. You can put the public key in source control so that
developers can add new secrets quickly and easily.

This renderer requires the gpg_ binary. No python libraries are required as of
the 2015.8.0 release.

.. _gpg-homedir:

GPG Homedir
-----------

The default `GPG Homedir <gpg-homedir>` is ``~/.gnupg`` and needs to be set using
``gpg --homedir``. Be very careful to not forget this option. It is also important
to run ``gpg`` commands as the user that owns the keys directory. If the salt-master
runs as user ``salt``, then use ``su - salt`` before running any gpg commands.

In some cases, it's preferable to have gpg keys stored on removable media or
other non-standard locations. This can be done using the ``gpg_keydir`` option
on the salt master. This will also require using a different path to ``--homedir``.

The ``--homedir`` argument can be configured for the current user using
``echo 'homedir /etc/salt/gpgkeys' >> ~/.gnupg``, but this should be used with
caution to avoid potential confusion.

.. code-block:: bash

    gpg_keydir: <path/to/homedir>

.. _gpg: https://gnupg.org

GPG Keys
--------

GPG key pairs include both a public and private key. The private key is akin to
a password and should be kept secure by the owner. A public key is used to
encrypt data being sent to the owner of the private key.

This means that the public key will be freely distributed so that others can
encrypt pillar data without access to the secret key.

New Key Pair
************

To create a new GPG key pair for encrypting data, log in to the master as root
and run the following:

.. code-block:: bash

    # mkdir -p /etc/salt/gpgkeys
    # chmod 0700 /etc/salt/gpgkeys
    # gpg --homedir /etc/salt/gpgkeys --gen-key

Do not supply a password for the keypair and use a name that makes sense for
your application.

.. note::
    In some situations, gpg may be starved of entropy and will take an incredibly
    long time to finish. Two common tools to generate (less secure) pseudo-random
    data are ``rng-tools`` and ``haveged``.

The new keys can be seen and verified using ``--list-secret-keys``:

.. code-block:: bash

    # gpg --homedir /etc/salt/gpgkeys --list-secret-keys
    /etc/salt/gpgkeys/pubring.kbx
    -----------------------------
    sec   rsa4096 2002-05-12 [SC] [expires: 2012-05-10]
          2DC47B416EE8C3484450B450A4D44406274AF44E
    uid           [ultimate] salt-master (gpg key for salt) <salt@cm.domain.tld>
    ssb   rsa4096 2002-05-12 [E] [expires: 2012-05-10]

In the example above, our KEY-ID is ``2DC47B416EE8C3484450B450A4D44406274AF44E``.

Export Public Key
*****************

To export a public key suitable for public distribution:

.. code-block:: bash

    # gpg --homedir /etc/salt/gpgkeys --armor --export <KEY-ID> > exported_pubkey.asc

.. _gpg-importpubkey:

Import Public Key
*****************

Users wishing to import the public key into their local keychain may run:

.. code-block:: bash

    $ gpg --import exported_pubkey.asc

Export (Save) Private Key
*************************

This key protects all gpg-encrypted pillar data and should be backed up to a
safe and secure location. This command will generate a backup of secret keys
in the ``/etc/salt/gpgkeys`` directory to the ``gpgkeys.secret`` file:

.. code-block:: bash

    # gpg --homedir /etc/salt/gpgkeys --export-secret-keys --export-options export-backup -o gpgkeys.secret

Salt does not support password-protected private keys, which means this file
is essentially a clear-text password (just add ``--armor``). Fortunately, it
is trivial to pass this export back to gpg to be encrypted with symmetric key:

.. code-block:: bash

    # gpg --homedir /etc/salt/gpgkeys --export-secret-keys --export-options export-backup | gpg --symmetric -o gpgkeys.gpg

.. note::
    In some cases, particularly when using su/sudo, gpg gets confused and needs
    to be told which TTY to use; this can be done with: ``export GPG_TTY=$(tty)``.

Import (Restore) Private Key
****************************

To import/restore a private key, create a directory with the correct permissions
and import using gpg.

.. code-block:: bash

    # mkdir -p /etc/salt/gpgkeys
    # chmod 0700 /etc/salt/gpgkeys
    # gpg --homedir /etc/salt/gpgkeys --import gpgkeys.secret

If the export was encrypted using a symmetric key, then decrypt first with:

.. code-block:: bash

    # gpg --decrypt gpgkeys.gpg | gpg --homedir /etc/salt/gpgkeys --import


Adjust trust level of imported keys
***********************************

In some cases, importing existing keys may not be enough and the trust level of
the key needs to be adjusted. This can be done by editing the key. The ``KEY-ID``
and the actual trust level of the key can be seen by listing the already imported
keys.


If the trust-level is not ``ultimate`` it needs to be changed by running

.. code-block:: bash

    gpg --homedir /etc/salt/gpgkeys --edit-key <KEY-ID>

This will open an interactive shell for the management of the GPG encryption key.
Type ``trust`` to be able to set the trust level for the key and then select ``5
(I trust ultimately)``. Then quit the shell by typing ``save``.

Encrypting Data
---------------

In order to encrypt data to a recipient (salt), the public key must be imported
into the local keyring. Importing the public key is described above in the
`Import Public Key <gpg-importpubkey:>` section.

To generate a cipher from a secret:

.. code-block:: bash

   $ echo -n 'supersecret' | gpg --trust-model always -ear <KEY-ID>

To apply the renderer on a file-by-file basis add the following line to the
top of any pillar with gpg data in it:

.. code-block:: yaml

    #!yaml|gpg

Now with your renderer configured, you can include your ciphers in your pillar
data like so:

.. code-block:: yaml

    #!yaml|gpg

    a-secret: |
      -----BEGIN PGP MESSAGE-----
      Version: GnuPG v1

      hQEMAweRHKaPCfNeAQf9GLTN16hCfXAbPwU6BbBK0unOc7i9/etGuVc5CyU9Q6um
      QuetdvQVLFO/HkrC4lgeNQdM6D9E8PKonMlgJPyUvC8ggxhj0/IPFEKmrsnv2k6+
      cnEfmVexS7o/U1VOVjoyUeliMCJlAz/30RXaME49Cpi6No2+vKD8a4q4nZN1UZcG
      RhkhC0S22zNxOXQ38TBkmtJcqxnqT6YWKTUsjVubW3bVC+u2HGqJHu79wmwuN8tz
      m4wBkfCAd8Eyo2jEnWQcM4TcXiF01XPL4z4g1/9AAxh+Q4d8RIRP4fbw7ct4nCJv
      Gr9v2DTF7HNigIMl4ivMIn9fp+EZurJNiQskLgNbktJGAeEKYkqX5iCuB1b693hJ
      FKlwHiJt5yA8X2dDtfk8/Ph1Jx2TwGS+lGjlZaNqp3R1xuAZzXzZMLyZDe5+i3RJ
      skqmFTbOiA===Eqsm
      -----END PGP MESSAGE-----

.. _encrypted-cli-pillar-data:

Encrypted CLI Pillar Data
-------------------------

.. versionadded:: 2016.3.0

Functions like :py:func:`state.highstate <salt.modules.state.highstate>` and
:py:func:`state.sls <salt.modules.state.sls>` allow for pillar data to be
passed on the CLI.

.. code-block:: bash

    salt myminion state.highstate pillar="{'mypillar': 'foo'}"

Starting with the 2016.3.0 release of Salt, it is now possible for this pillar
data to be GPG-encrypted, and to use the GPG renderer to decrypt it.

Replacing Newlines
******************

To pass encrypted pillar data on the CLI, the ciphertext must have its newlines
replaced with a literal backslash-n (``\n``), as newlines are not supported
within Salt CLI arguments. There are a number of ways to do this:

With awk or Perl:

.. code-block:: bash

    # awk
    ciphertext=`echo -n "supersecret" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | awk '{printf "%s\\n",$0} END {print ""}'`
    # Perl
    ciphertext=`echo -n "supersecret" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | perl -pe 's/\n/\\n/g'`

With Python:

.. code-block:: python

    import subprocess

    secret, stderr = subprocess.Popen(
        ['gpg', '--armor', '--batch', '--trust-model', 'always', '--encrypt',
         '-r', 'user@domain.com'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE).communicate(input='supersecret')

    if secret:
        print(secret.replace('\n', r'\n'))
    else:
        raise ValueError('No ciphertext found: {0}'.format(stderr))

.. code-block:: bash

    ciphertext=`python /path/to/script.py`


The ciphertext can be included in the CLI pillar data like so:

.. code-block:: bash

    salt myminion state.sls secretstuff pillar_enc=gpg pillar="{secret_pillar: '$ciphertext'}"

The ``pillar_enc=gpg`` argument tells Salt that there is GPG-encrypted pillar
data, so that the CLI pillar data is passed through the GPG renderer, which
will iterate recursively though the CLI pillar dictionary to decrypt any
encrypted values.


Encrypting the Entire CLI Pillar Dictionary
*******************************************

If several values need to be encrypted, it may be more convenient to encrypt
the entire CLI pillar dictionary. Again, this can be done in several ways:

With awk or Perl:

.. code-block:: bash

    # awk
    ciphertext=`echo -n "{'secret_a': 'CorrectHorseBatteryStaple', 'secret_b': 'GPG is fun!'}" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | awk '{printf "%s\\n",$0} END {print ""}'`
    # Perl
    ciphertext=`echo -n "{'secret_a': 'CorrectHorseBatteryStaple', 'secret_b': 'GPG is fun!'}" | gpg --armor --batch --trust-model always --encrypt -r user@domain.com | perl -pe 's/\n/\\n/g'`

With Python:

.. code-block:: python

    import subprocess

    pillar_data = {'secret_a': 'CorrectHorseBatteryStaple',
                   'secret_b': 'GPG is fun!'}

    secret, stderr = subprocess.Popen(
        ['gpg', '--armor', '--batch', '--trust-model', 'always', '--encrypt',
         '-r', 'user@domain.com'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE).communicate(input=repr(pillar_data))

    if secret:
        print(secret.replace('\n', r'\n'))
    else:
        raise ValueError('No ciphertext found: {0}'.format(stderr))

.. code-block:: bash

    ciphertext=`python /path/to/script.py`

With the entire pillar dictionary now encrypted, it can be included in the CLI
pillar data like so:

.. code-block:: bash

    salt myminion state.sls secretstuff pillar_enc=gpg pillar="$ciphertext"

Configuration
*************

The default behaviour of this renderer is to log a warning if a block could not
be decrypted; in other words, it just returns the ciphertext rather than the
encrypted secret.

This behaviour can be changed via the `gpg_decrypt_must_succeed` configuration
option.  If set to `True`, any gpg block that cannot be decrypted raises a
`SaltRenderError` exception, which registers an error in ``_errors`` during
rendering.

In the Chlorine release, the default behavior will be reversed and an error
message will be added to ``_errors`` by default.
"""


import logging
import os
import re
from subprocess import PIPE, Popen

import salt.syspaths
import salt.utils.cache
import salt.utils.path
import salt.utils.stringio
import salt.utils.stringutils
import salt.utils.versions
from salt.exceptions import SaltRenderError

log = logging.getLogger(__name__)

GPG_CIPHERTEXT = re.compile(
    salt.utils.stringutils.to_bytes(
        r"-----BEGIN PGP MESSAGE-----.*?-----END PGP MESSAGE-----"
    ),
    re.DOTALL,
)
GPG_CACHE = None


def _get_gpg_exec():
    """
    return the GPG executable or raise an error
    """
    gpg_exec = salt.utils.path.which("gpg")
    if gpg_exec:
        return gpg_exec
    else:
        raise SaltRenderError("GPG unavailable")


def _get_key_dir():
    """
    return the location of the GPG key directory
    """
    gpg_keydir = None
    if "config.get" in __salt__:
        gpg_keydir = __salt__["config.get"]("gpg_keydir")

    if not gpg_keydir:
        gpg_keydir = __opts__.get(
            "gpg_keydir",
            os.path.join(
                __opts__.get("config_dir", os.path.dirname(__opts__["conf_file"])),
                "gpgkeys",
            ),
        )

    return gpg_keydir


def _get_cache():
    global GPG_CACHE
    if not GPG_CACHE:
        cachedir = __opts__.get("cachedir")
        GPG_CACHE = salt.utils.cache.CacheFactory.factory(
            __opts__.get("gpg_cache_backend"),
            __opts__.get("gpg_cache_ttl"),
            minion_cache_path=os.path.join(cachedir, "gpg_cache"),
        )
    return GPG_CACHE


def _decrypt_ciphertext(cipher):
    """
    Given a block of ciphertext as a string, and a gpg object, try to decrypt
    the cipher and return the decrypted string. If the cipher cannot be
    decrypted, log the error, and return the ciphertext back out.
    """
    try:
        cipher = salt.utils.stringutils.to_unicode(cipher).replace(r"\n", "\n")
    except UnicodeDecodeError:
        # ciphertext is binary
        pass
    cipher = salt.utils.stringutils.to_bytes(cipher)
    if __opts__.get("gpg_cache"):
        cache = _get_cache()
        if cipher in cache:
            return cache[cipher]
    cmd = [
        _get_gpg_exec(),
        "--homedir",
        _get_key_dir(),
        "--status-fd",
        "2",
        "--no-tty",
        "-d",
    ]
    proc = Popen(cmd, stdin=PIPE, stdout=PIPE, stderr=PIPE, shell=False)
    decrypted_data, decrypt_error = proc.communicate(input=cipher)
    if not decrypted_data:
        log.warning("Could not decrypt cipher %r, received: %r", cipher, decrypt_error)
        if __opts__["gpg_decrypt_must_succeed"]:
            raise SaltRenderError(
                "Could not decrypt cipher {!r}, received: {!r}".format(
                    cipher,
                    decrypt_error,
                )
            )
        else:
            salt.utils.versions.warn_until(
                "Chlorine",
                "After the Chlorine release of Salt, gpg_decrypt_must_succeed will default to True.",
            )
        return cipher
    else:
        if __opts__.get("gpg_cache"):
            cache[cipher] = decrypted_data
        return decrypted_data


def _decrypt_ciphertexts(cipher, translate_newlines=False, encoding=None):
    to_bytes = salt.utils.stringutils.to_bytes
    cipher = to_bytes(cipher)
    if translate_newlines:
        cipher = cipher.replace(to_bytes(r"\n"), to_bytes("\n"))

    def replace(match):
        result = to_bytes(_decrypt_ciphertext(match.group()))
        return result

    ret, num = GPG_CIPHERTEXT.subn(replace, to_bytes(cipher))
    if num > 0:
        # Remove trailing newlines. Without if crypted value initially specified as a YAML multiline
        # it will conain unexpected trailing newline.
        ret = ret.rstrip(b"\n")
    else:
        ret = cipher

    try:
        ret = salt.utils.stringutils.to_unicode(ret, encoding=encoding)
    except UnicodeDecodeError:
        # decrypted data contains some sort of binary data - not our problem
        pass
    return ret


def _decrypt_object(obj, translate_newlines=False, encoding=None):
    """
    Recursively try to decrypt any object. If the object is a string
    or bytes and it contains a valid GPG header, decrypt it,
    otherwise keep going until a string is found.
    """
    if salt.utils.stringio.is_readable(obj):
        return _decrypt_object(obj.getvalue(), translate_newlines)
    if isinstance(obj, (str, bytes)):
        return _decrypt_ciphertexts(
            obj, translate_newlines=translate_newlines, encoding=encoding
        )
    elif isinstance(obj, dict):
        for key, value in obj.items():
            obj[key] = _decrypt_object(value, translate_newlines=translate_newlines)
        return obj
    elif isinstance(obj, list):
        for key, value in enumerate(obj):
            obj[key] = _decrypt_object(value, translate_newlines=translate_newlines)
        return obj
    else:
        return obj


def render(gpg_data, saltenv="base", sls="", argline="", **kwargs):
    """
    Create a gpg object given a gpg_keydir, and then use it to try to decrypt
    the data to be rendered.
    """
    if not _get_gpg_exec():
        raise SaltRenderError("GPG unavailable")
    log.debug("Reading GPG keys from: %s", _get_key_dir())

    translate_newlines = kwargs.get("translate_newlines", False)
    return _decrypt_object(
        gpg_data,
        translate_newlines=translate_newlines,
        encoding=kwargs.get("encoding", None),
    )



"""
Module for running nictagadm command on SmartOS
:maintainer:    Jorge Schrauwen <sjorge@blackdot.be>
:maturity:      new
:depends:       nictagadm binary, dladm binary
:platform:      smartos

.. versionadded:: 2016.11.0

"""

import logging

import salt.utils.path
import salt.utils.platform

log = logging.getLogger(__name__)

# Function aliases
__func_alias__ = {"list_nictags": "list"}

# Define the module's virtual name
__virtualname__ = "nictagadm"


def __virtual__():
    """
    Provides nictagadm on SmartOS
    """
    if (
        salt.utils.platform.is_smartos_globalzone()
        and salt.utils.path.which("dladm")
        and salt.utils.path.which("nictagadm")
    ):
        return __virtualname__
    return (
        False,
        "{} module can only be loaded on SmartOS compute nodes".format(__virtualname__),
    )


def list_nictags(include_etherstubs=True):
    """
    List all nictags

    include_etherstubs : boolean
        toggle include of etherstubs

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.list
    """
    ret = {}
    cmd = 'nictagadm list -d "|" -p{}'.format(" -L" if not include_etherstubs else "")
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = (
            res["stderr"] if "stderr" in res else "Failed to get list of nictags."
        )
    else:
        header = ["name", "macaddress", "link", "type"]
        for nictag in res["stdout"].splitlines():
            nictag = nictag.split("|")
            nictag_data = {}
            for field in header:
                nictag_data[field] = nictag[header.index(field)]
            ret[nictag_data["name"]] = nictag_data
            del ret[nictag_data["name"]]["name"]
    return ret


def vms(nictag):
    """
    List all vms connect to nictag

    nictag : string
        name of nictag

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.vms admin
    """
    ret = {}
    cmd = "nictagadm vms {}".format(nictag)
    res = __salt__["cmd.run_all"](cmd)
    retcode = res["retcode"]
    if retcode != 0:
        ret["Error"] = (
            res["stderr"] if "stderr" in res else "Failed to get list of vms."
        )
    else:
        ret = res["stdout"].splitlines()
    return ret


def exists(*nictag, **kwargs):
    """
    Check if nictags exists

    nictag : string
        one or more nictags to check
    verbose : boolean
        return list of nictags

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.exists admin
    """
    ret = {}
    if not nictag:
        return {"Error": "Please provide at least one nictag to check."}

    cmd = "nictagadm exists -l {}".format(" ".join(nictag))
    res = __salt__["cmd.run_all"](cmd)

    if not kwargs.get("verbose", False):
        ret = res["retcode"] == 0
    else:
        missing = res["stderr"].splitlines()
        for nt in nictag:
            ret[nt] = nt not in missing

    return ret


def add(name, mac, mtu=1500):
    """
    Add a new nictag

    name : string
        name of new nictag
    mac : string
        mac of parent interface or 'etherstub' to create a ether stub
    mtu : int
        MTU (ignored for etherstubs)

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.add storage0 etherstub
        salt '*' nictagadm.add trunk0 'DE:AD:OO:OO:BE:EF' 9000
    """
    ret = {}

    if mtu > 9000 or mtu < 1500:
        return {"Error": "mtu must be a value between 1500 and 9000."}
    if mac != "etherstub":
        cmd = "dladm show-phys -m -p -o address"
        res = __salt__["cmd.run_all"](cmd)
        # dladm prints '00' as '0', so account for that.
        if mac.replace("00", "0") not in res["stdout"].splitlines():
            return {"Error": "{} is not present on this system.".format(mac)}

    if mac == "etherstub":
        cmd = "nictagadm add -l {}".format(name)
        res = __salt__["cmd.run_all"](cmd)
    else:
        cmd = "nictagadm add -p mtu={},mac={} {}".format(mtu, mac, name)
        res = __salt__["cmd.run_all"](cmd)

    if res["retcode"] == 0:
        return True
    else:
        return {
            "Error": "failed to create nictag."
            if "stderr" not in res and res["stderr"] == ""
            else res["stderr"]
        }


def update(name, mac=None, mtu=None):
    """
    Update a nictag

    name : string
        name of nictag
    mac : string
        optional new mac for nictag
    mtu : int
        optional new MTU for nictag

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.update trunk mtu=9000
    """
    ret = {}

    if name not in list_nictags():
        return {"Error": "nictag {} does not exists.".format(name)}
    if not mtu and not mac:
        return {"Error": "please provide either mac or/and mtu."}
    if mtu:
        if mtu > 9000 or mtu < 1500:
            return {"Error": "mtu must be a value between 1500 and 9000."}
    if mac:
        if mac == "etherstub":
            return {"Error": 'cannot update a nic with "etherstub".'}
        else:
            cmd = "dladm show-phys -m -p -o address"
            res = __salt__["cmd.run_all"](cmd)
            # dladm prints '00' as '0', so account for that.
            if mac.replace("00", "0") not in res["stdout"].splitlines():
                return {"Error": "{} is not present on this system.".format(mac)}

    if mac and mtu:
        properties = "mtu={},mac={}".format(mtu, mac)
    elif mac:
        properties = "mac={}".format(mac) if mac else ""
    elif mtu:
        properties = "mtu={}".format(mtu) if mtu else ""

    cmd = "nictagadm update -p {} {}".format(properties, name)
    res = __salt__["cmd.run_all"](cmd)

    if res["retcode"] == 0:
        return True
    else:
        return {
            "Error": "failed to update nictag."
            if "stderr" not in res and res["stderr"] == ""
            else res["stderr"]
        }


def delete(name, force=False):
    """
    Delete nictag

    name : string
        nictag to delete
    force : boolean
        force delete even if vms attached

    CLI Example:

    .. code-block:: bash

        salt '*' nictagadm.exists admin
    """
    ret = {}

    if name not in list_nictags():
        return True

    cmd = "nictagadm delete {}{}".format("-f " if force else "", name)
    res = __salt__["cmd.run_all"](cmd)

    if res["retcode"] == 0:
        return True
    else:
        return {
            "Error": "failed to delete nictag."
            if "stderr" not in res and res["stderr"] == ""
            else res["stderr"]
        }



"""
Manage S3 Resources
===================

.. versionadded:: 2018.3.0

Manage S3 resources. Be aware that this interacts with Amazon's services,
and so may incur charges.

This module uses ``boto3``, which can be installed via package, or pip.

This module accepts explicit AWS credentials but can also utilize
IAM roles assigned to the instance through Instance Profiles. Dynamic
credentials are then automatically obtained from AWS API and no further
configuration is necessary. More information available `here
<http://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html>`_.

If IAM roles are not used you need to specify them either in a pillar file or
in the minion's config file:

.. code-block:: yaml

    s3.keyid: GKTADJGHEIQSXMKKRBJ08H
    s3.key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs

It's also possible to specify ``key``, ``keyid`` and ``region`` via a profile,
either passed in as a dict, or as a string to pull from pillars or minion
config:

.. code-block:: yaml

    myprofile:
        keyid: GKTADJGHEIQSXMKKRBJ08H
        key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            region: us-east-1

.. code-block:: yaml

    Ensure s3 object exists:
        boto_s3.object_present:
            - name: s3-bucket/s3-key
            - source: /path/to/local/file
            - region: us-east-1
            - keyid: GKTADJGHEIQSXMKKRBJ08H
            - key: askdjghsdfjkghWupUjasdflkdfklgjsdfjajkghs
            - profile: my-profile

:depends: boto3
"""


import copy
import difflib
import logging

import salt.utils.hashutils

log = logging.getLogger(__name__)


def __virtual__():
    """
    Only load if boto is available.
    """
    if "boto_s3.get_object_metadata" not in __salt__:
        return (False, "boto_s3 module could not be loaded")
    return "boto_s3"


# Keys for `extra_args` that we support.
# Currently, this excludes the `ACL` and `Grant*` keys.
# Most keys are stored and returned by AWS:
STORED_EXTRA_ARGS = frozenset(
    [
        "CacheControl",
        "ContentDisposition",
        "ContentEncoding",
        "ContentLanguage",
        "ContentType",
        "Expires",
        "Metadata",
        "ServerSideEncryption",
        "SSECustomerAlgorithm",
        "SSECustomerKeyMD5",
        "SSEKMSKeyId",
        "StorageClass",
        "WebsiteRedirectLocation",
    ]
)
# However, some keys are only specified on upload,
# but won't be stored/returned by AWS as metadata:
UPLOAD_ONLY_EXTRA_ARGS = frozenset(
    [
        # AWS doesn't store customer provided keys,
        # can use SSECustomerKeyMD5 to check for correct key
        "SSECustomerKey",
        "RequestPayer",
    ]
)
# Some extra args must also be passed along to retrive metadata,
# namely SSE-C (customer-provided encryption) and RequestPayer args.
GET_METADATA_EXTRA_ARGS = frozenset(
    ["SSECustomerAlgorithm", "SSECustomerKey", "SSECustomerKeyMD5", "RequestPayer"]
)


def object_present(
    name,
    source=None,
    hash_type=None,
    extra_args=None,
    extra_args_from_pillar="boto_s3_object_extra_args",
    region=None,
    key=None,
    keyid=None,
    profile=None,
):
    """
    Ensure object exists in S3.

    name
        The name of the state definition.
        This will be used to determine the location of the object in S3,
        by splitting on the first slash and using the first part
        as the bucket name and the remainder as the S3 key.

    source
        The source file to upload to S3,
        currently this only supports files hosted on the minion's local
        file system (starting with /).

    hash_type
        Hash algorithm to use to check that the object contents are correct.
        Defaults to the value of the `hash_type` config option.

    extra_args
        A dictionary of extra arguments to use when uploading the file.
        Note that these are only enforced if new objects are uploaded,
        and not modified on existing objects.
        The supported args are those in the ALLOWED_UPLOAD_ARGS list at
        http://boto3.readthedocs.io/en/latest/reference/customizations/s3.html.
        However, Note that the 'ACL', 'GrantFullControl', 'GrantRead',
        'GrantReadACP',  and 'GrantWriteACL' keys are currently not supported.

    extra_args_from_pillar
        Name of pillar dict that contains extra arguments.
        Extra arguments defined for this specific state will be
        merged over those from the pillar.

    region
        Region to connect to.

    key
        Secret key to be used.

    keyid
        Access key to be used.

    profile
        A dict with region, key and keyid, or a pillar key (string) that
        contains a dict with region, key and keyid.
    """
    ret = {
        "name": name,
        "comment": "",
        "changes": {},
    }

    if extra_args is None:
        extra_args = {}
    combined_extra_args = copy.deepcopy(
        __salt__["config.option"](extra_args_from_pillar, {})
    )
    __utils__["dictupdate.update"](combined_extra_args, extra_args)
    if combined_extra_args:
        supported_args = STORED_EXTRA_ARGS | UPLOAD_ONLY_EXTRA_ARGS
        combined_extra_args_keys = frozenset(combined_extra_args.keys())
        extra_keys = combined_extra_args_keys - supported_args
        if extra_keys:
            msg = "extra_args keys {} are not supported".format(extra_keys)
            return {"error": msg}

    # Get the hash of the local file
    if not hash_type:
        hash_type = __opts__["hash_type"]
    try:
        digest = salt.utils.hashutils.get_hash(source, form=hash_type)
    except OSError as e:
        ret["result"] = False
        ret["comment"] = "Could not read local file {}: {}".format(
            source,
            e,
        )
        return ret
    except ValueError as e:
        # Invalid hash type exception from get_hash
        ret["result"] = False
        ret["comment"] = "Could not hash local file {}: {}".format(
            source,
            e,
        )
        return ret

    HASH_METADATA_KEY = "salt_managed_content_hash"
    combined_extra_args.setdefault("Metadata", {})
    if HASH_METADATA_KEY in combined_extra_args["Metadata"]:
        # Be lenient, silently allow hash metadata key if digest value matches
        if combined_extra_args["Metadata"][HASH_METADATA_KEY] != digest:
            ret["result"] = False
            ret["comment"] = (
                "Salt uses the {} metadata key internally,"
                "do not pass it to the boto_s3.object_present state.".format(
                    HASH_METADATA_KEY
                )
            )
            return ret
    combined_extra_args["Metadata"][HASH_METADATA_KEY] = digest
    # Remove upload-only keys from full set of extra_args
    # to create desired dict for comparisons
    desired_metadata = {
        k: v for k, v in combined_extra_args.items() if k not in UPLOAD_ONLY_EXTRA_ARGS
    }

    # Some args (SSE-C, RequestPayer) must also be passed to get_metadata
    metadata_extra_args = {
        k: v for k, v in combined_extra_args.items() if k in GET_METADATA_EXTRA_ARGS
    }
    r = __salt__["boto_s3.get_object_metadata"](
        name,
        extra_args=metadata_extra_args,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )
    if "error" in r:
        ret["result"] = False
        ret["comment"] = "Failed to check if S3 object exists: {}.".format(
            r["error"],
        )
        return ret

    if r["result"]:
        # Check if content and metadata match
        # A hash of the content is injected into the metadata,
        # so we can combine both checks into one
        # Only check metadata keys specified by the user,
        # ignore other fields that have been set
        s3_metadata = {
            k: r["result"][k]
            for k in STORED_EXTRA_ARGS
            if k in desired_metadata and k in r["result"]
        }
        if s3_metadata == desired_metadata:
            ret["result"] = True
            ret["comment"] = "S3 object {} is present.".format(name)
            return ret
        action = "update"
    else:
        s3_metadata = None
        action = "create"

    def _yaml_safe_dump(attrs):
        """
        Safely dump YAML using a readable flow style
        """
        dumper_name = "IndentedSafeOrderedDumper"
        dumper = __utils__["yaml.get_dumper"](dumper_name)
        return __utils__["yaml.dump"](attrs, default_flow_style=False, Dumper=dumper)

    changes_diff = "".join(
        difflib.unified_diff(
            _yaml_safe_dump(s3_metadata).splitlines(True),
            _yaml_safe_dump(desired_metadata).splitlines(True),
        )
    )

    if __opts__["test"]:
        ret["result"] = None
        ret["comment"] = "S3 object {} set to be {}d.".format(name, action)
        ret["comment"] += "\nChanges:\n{}".format(changes_diff)
        ret["changes"] = {"diff": changes_diff}
        return ret

    r = __salt__["boto_s3.upload_file"](
        source,
        name,
        extra_args=combined_extra_args,
        region=region,
        key=key,
        keyid=keyid,
        profile=profile,
    )

    if "error" in r:
        ret["result"] = False
        ret["comment"] = "Failed to {} S3 object: {}.".format(
            action,
            r["error"],
        )
        return ret

    ret["result"] = True
    ret["comment"] = "S3 object {} {}d.".format(name, action)
    ret["comment"] += "\nChanges:\n{}".format(changes_diff)
    ret["changes"] = {"diff": changes_diff}
    return ret
